# -*- coding: utf-8 -*-
from .server import Server
import requests, time, json, rsa, binascii, base64
import httpx

from .services.TalkService import TalkService
from .services.ShopService import ShopService
from .services.LiffService import LiffService
from .services.ChannelService import ChannelService
from .services.SquareService import SquareService
from .services.BuddyService import BuddyService
from .services.PrimaryAccountInitService import PrimaryAccountInitService
from .services.AuthService import AuthService
from .services.SettingsService import SettingsService
from .services.AccessTokenRefreshService import AccessTokenRefreshService
from .services.CallService import CallService
from .services.SecondaryPwlessLoginService import SecondaryPwlessLoginService
from .services.SecondaryPwlessLoginPermitNoticeService import SecondaryPwlessLoginPermitNoticeService

class API(TalkService, ShopService, LiffService, ChannelService, SquareService, BuddyService, PrimaryAccountInitService, AuthService, SettingsService, AccessTokenRefreshService, CallService, SecondaryPwlessLoginService, SecondaryPwlessLoginPermitNoticeService):
    _msgSeq = 0
    url = "https://gf.line.naver.jp/enc"
    
    def __init__(self):
        self.server = Server()
        self.req = requests.session()
        self.req_h2 = httpx.Client(http2=True)
        self.server.Headers = {
            "x-line-application": self.APP_NAME,
            "x-le": self.le,
            "x-lap": "5",
            # "x-lc": "3",
            # "X-LST": "",
            # "X-LCR": "",
            # "X-LOR": "",
            "x-lpv": "1",
            "x-lcs": self._encryptKey,
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36",
            "content-type": "application/x-thrift; protocol=TBINARY",
            "x-lal": self.LINE_LANGUAGE,
            "x-lhm": "POST"
        }
        self.authToken = None
        self.revision = 0
        self.globalRev = 0
        self.individualRev = 0
        self._msgSeq = 0
        TalkService.__init__(self)
        PrimaryAccountInitService.__init__(self)
        AuthService.__init__(self)
        SettingsService.__init__(self)
        AccessTokenRefreshService.__init__(self)
        CallService.__init__(self)
        SecondaryPwlessLoginService.__init__(self)
        SecondaryPwlessLoginPermitNoticeService.__init__(self)
    
    def requestPwlessLogin(self, phone, pw):
        pwless_code = self.createPwlessSession(phone)
        pwless_code = pwless_code[1]
        print(f'PWLESS SESSION: {pwless_code}')
        certVerify = self.verifyLoginCertificate(pwless_code, self.getCacheData('.pwless', phone))
        if 'error' in certVerify:
            pwless_pincode = self.requestPinCodeVerif(pwless_code)[1]
            print(f'PWLESS PINCODE: {pwless_pincode}')
            certVerify = self.checkPwlessPinCodeVerified(pwless_code)
        if certVerify is not None and 'error' not in certVerify:
            secret, secretPK = self.createSqrSecret(True)
            self.putExchangeKey(pwless_code, secretPK)
            self.requestPaakAuth(pwless_code)
            print(f'need Paak Auth Confind')
            pa = self.checkPaakAuthenticated(pwless_code)
            if pa is not None and 'error' not in pa:
                ek = self.getE2eeKey(pwless_code)
                loginInfo = self.pwlessLoginV2(pwless_code)
                if 'error' not in loginInfo:
                    cert = loginInfo[2]
                    tokenInfo = loginInfo[3]
                    token = tokenInfo[1]
                    token2 = tokenInfo[2]
                    mid = loginInfo[5]
                else:
                    loginInfo = self.pwlessLogin(pwless_code)
                    token = loginInfo[1]
                    cert = loginInfo[2]
                    mid = loginInfo[4]
                self.authToken = token
                self.decodeE2EEKeyV1(ek[1], secret, mid)
                self.saveCacheData('.pwless', phone, cert)
                print(f'Auth Token: {self.authToken}')
                return True
        raise Exception('login failed.')

    def requestEmailLogin(self, email, pw):
        rsaKey = self.getRSAKeyInfo()
        keynm = rsaKey[1]
        nvalue = rsaKey[2]
        evalue = rsaKey[3]
        sessionKey = rsaKey[4]
        certificate = self.getEmailCert(email)
        message = (chr(len(sessionKey)) + sessionKey +
           chr(len(email)) + email +
           chr(len(pw)) + pw).encode('utf-8')
        pub_key = rsa.PublicKey(int(nvalue, 16), int(evalue, 16))
        crypto = binascii.hexlify(rsa.encrypt(message, pub_key)).decode()
        res = self.loginZ(keynm, crypto, certificate=certificate)
        if 1 not in res:
            print(f"Enter Pincode: {res[4]}")
            verifier = self.checkLoginZPinCode(res[3])['verifier']
            res = self.loginZ(keynm, crypto, verifier=verifier)
            self.saveEmailCert(email, res[2])
        self.authToken = res[1]
        print(f"AuthToken: {self.authToken}")
        return True

    def requestEmailLoginV2(self, email, pw):
        rsaKey = self.getRSAKeyInfo()
        keynm = rsaKey[1]
        nvalue = rsaKey[2]
        evalue = rsaKey[3]
        sessionKey = rsaKey[4]
        certificate = self.getEmailCert(email)
        message = (chr(len(sessionKey)) + sessionKey +
           chr(len(email)) + email +
           chr(len(pw)) + pw).encode('utf-8')
        pub_key = rsa.PublicKey(int(nvalue, 16), int(evalue, 16))
        crypto = binascii.hexlify(rsa.encrypt(message, pub_key)).decode()
        secret, secretPK = self.createSqrSecret(True)
        pincode = b"1314520"
        _secret = self._encryptAESECB(self.getSHA256Sum(pincode), base64.b64decode(secretPK))
        res = self.loginV2(keynm, crypto, _secret, deviceName=self.SYSTEM_NAME, cert=certificate)
        if res.get('error', {}).get('code', -1) == 20:
            print(f"can't login: {res['error']['message']}, try use LoginZ...")
            return self.requestEmailLogin(email, pw)
        if 9 not in res:
            verifier = res[3]
            if res[5] == 3:
                print(f'need device confirm')
            print(f"Enter Pincode: {pincode.decode()}")
            e2eeInfo = self.checkLoginV2PinCode(verifier)['metadata']
            try:
                e2eeKeyInfo = self.decodeE2EEKeyV1(e2eeInfo, secret)
            except:
                raise Exception(f"e2eeInfo decode failed, try again")
            blablabao = self.encryptDeviceSecret(base64.b64decode(e2eeInfo['publicKey']), secret, base64.b64decode(e2eeInfo['encryptedKeyChain']))
            e2eeLogin = self.confirmE2EELogin(verifier, blablabao)
            if 'error' not in e2eeLogin:
                res = self.loginV2(None, None, None, deviceName=self.SYSTEM_NAME, verifier=e2eeLogin)
                if res.get('error', {}).get('code', -1) == 20:
                    print(f"can't login: {res['error']['message']}, try use LoginZ...")
                    return self.requestEmailLogin(email, pw)
                self.saveEmailCert(email, res[2])
            else:
                raise Exception(f"confirmE2EELogin failed, try again")
        self.authToken = res[9][1]
        refreshToken = res[9][2]
        self.saveCacheData('.refreshToken', self.authToken, refreshToken)
        print(f"AuthToken: {self.authToken}")
        print(f"RefreshToken: {refreshToken}")
        return True

    def requestSQR(self, isSelf=True):
        sqr = self.createSession()[1]
        url = self.createQrCode(sqr)[1]
        secret, secretUrl = self.createSqrSecret()
        yield f"URL: {url}{secretUrl}"
        if self.checkQrCodeVerified(sqr):
            b = self.verifyCertificate(sqr, self.getSqrCert())
            isCheck = False
            if b is not None and 'error' in b:
                c = self.createPinCode(sqr)
                yield f"請輸入pincode: {c}"
                if self.checkPinCodeVerified(sqr):
                    isCheck = True
            else:
                isCheck = True
            if isCheck:
                e = self.qrCodeLogin(sqr, secret)
                print(e)
                if isSelf:
                    self.authToken = e
                    print(f"AuthToken: {self.authToken}")
                else:
                    yield e
                    return
                yield self.authToken
                return
            raise Exception('can not check pin code, try again?')
        raise Exception('can not check qr code, try again?')
        
    def createSession(self):
        params = []
        sqrd = self.generateDummyProtocol('createSession', params, 3)
        return self.postPackDataAndGetUnpackRespData('/acct/lgn/sq/v1' ,sqrd, 3)
    
    def createQrCode(self, qrcode):
        params = [
            [12, 1, [
                [11, 1, qrcode]
            ]]
        ]
        sqrd = self.generateDummyProtocol('createQrCode', params, 3)
        return self.postPackDataAndGetUnpackRespData('/acct/lgn/sq/v1' ,sqrd, 3)
        
    def checkQrCodeVerified(self, qrcode):
        _headers = {
            "X-Line-Access": qrcode,
            "x-lpqs": "/acct/lp/lgn/sq/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 19, 99, 104, 101, 99, 107, 81, 114, 67, 111, 100, 101, 86, 101, 114, 105, 102, 105, 101, 100, 0, 0, 0, 0, 12, 0, 1, 11, 0, 1, 0, 0, 0, 66]
        for qr in qrcode:
            sqrd.append(ord(qr))
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        if res.status_code == 200:
            return True
        return False
        
    def verifyCertificate(self, qrcode, cert=None):
        _headers = {
            "x-lpqs": "/acct/lgn/sq/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 17, 118, 101, 114, 105, 102, 121, 67, 101, 114, 116, 105, 102, 105, 99, 97, 116, 101, 0, 0, 0, 0, 12, 0, 1, 11, 0, 1, 0, 0, 0, 66]
        for qr in qrcode:
            sqrd.append(ord(qr))
        if cert is not None:
            sqrd += [11, 0, 2] + self.getStringBytes(cert)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)

    def createPinCode(self, qrcode):
        _headers = {
            "x-lpqs": "/acct/lgn/sq/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 13, 99, 114, 101, 97, 116, 101, 80, 105, 110, 67, 111, 100, 101, 0, 0, 0, 0, 12, 0, 1, 11, 0, 1, 0, 0, 0, 66]
        for qr in qrcode:
            sqrd.append(ord(qr))
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return data[39:43].decode()
        
    def checkPinCodeVerified(self, qrcode):
        _headers = {
            "X-Line-Access": qrcode,
            "x-lpqs": "/acct/lp/lgn/sq/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 20, 99, 104, 101, 99, 107, 80, 105, 110, 67, 111, 100, 101, 86, 101, 114, 105, 102, 105, 101, 100, 0, 0, 0, 0, 12, 0, 1, 11, 0, 1, 0, 0, 0, 66]
        for qr in qrcode:
            sqrd.append(ord(qr))
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        if res.status_code == 200:
            return True
        return False
        
    def qrCodeLogin(self, qrcode, secret):
        _headers = {
            "x-lpqs": "/acct/lgn/sq/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 11, 113, 114, 67, 111, 100, 101, 76, 111, 103, 105, 110, 0, 0, 0, 0, 12, 0, 1, 11, 0, 1, 0, 0, 0, 66]
        for qr in qrcode:
            sqrd.append(ord(qr))
        self.APP_TYPE = 'CHANNELGW'
        sqrd += [11, 0, 2, 0, 0, 0, len(self.APP_TYPE)]
        for device in self.APP_TYPE:
            sqrd.append(ord(device))
        sqrd += [2, 0, 3, 0, 0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        data = self.tryReadData(data)
        pem = data[1]
        self.saveSqrCert(pem)
        print("證書: ", pem)
        _mid = data[5]
        if data.get(4) is not None:
            self.decodeE2EEKeyV1(data[4], secret, _mid)
        _token = data[2]
        return _token
    
    def qrCodeLoginV2(self, authSessionId, systemName="彥彥好睡",modelName="鴻鴻好暈", autoLoginIsRequired=True):
        params = [
            [12, 1, [
                [11, 1, authSessionId],
                [11, 2, self.APP_TYPE],
                [11, 3, self.APP_TYPE],
                [2, 4, autoLoginIsRequired]
            ]]
        ]
        sqrd = self.generateDummyProtocol('qrCodeLoginV2', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/lgn/sq/v1" ,sqrd, 3)
        
    def CPF(self):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/CPF"
        }
        a = self.encHeaders(_headers)
        sqrd = []
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return bytes(data)
        
    def returnTicket(self, searchId, fromEnvInfo, otp):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/S3" # V3?
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 12, 114, 101, 116, 117, 114, 110, 84, 105, 99, 107, 101, 116, 0, 0, 0, 0]
        sqrd += [12, 0, 1] # AcquireOACallRouteRequest
        sqrd += [11, 0, 1, 0, 0, 0, len(searchId)]
        for value in searchId:
            sqrd.append(ord(value))
        # sqrd += [13, 0, 2, 0, 0, 0, len(otp)] #todo?
        sqrd += [11, 0, 3, 0, 0, 0, len(otp)]
        for value in otp:
            sqrd.append(ord(value))
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def wakeUpLongPolling(self, clientRevision):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/P3" # P3? S3?
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 17, 119, 97, 107, 101, 85, 112, 76, 111, 110, 103, 80, 111, 108, 108, 105, 110, 103, 0, 0, 0, 0]
        sqrd += [10, 0, 2] + self.getIntBytes(clientRevision, 8)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def getModulesV2(self, etag):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/WALLET3"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 12, 103, 101, 116, 77, 111, 100, 117, 108, 101, 115, 86, 50, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [11, 0, 1, 0, 0, 0, len(etag)] # etag
        for value in etag:
            sqrd.append(ord(value))
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def openSession(self, udid, deviceModel):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/ACCT/authfactor/eap/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 11, 111, 112, 101, 110, 83, 101, 115, 115, 105, 111, 110, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [12, 0, 1]
        udid = str(udid).encode()
        sqrd += [11, 0, 1] + self.getIntBytes(len(udid))
        for value2 in udid:
            sqrd.append(value2)
        deviceModel = str(deviceModel).encode()
        sqrd += [11, 0, 2] + self.getIntBytes(len(deviceModel))
        for value2 in deviceModel:
            sqrd.append(value2)
        sqrd += [0, 0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def connectEapAccount(self, authSessionId):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/ACCT/authfactor/eap/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 17, 99, 111, 110, 110, 101, 99, 116, 69, 97, 112, 65, 99, 99, 111, 117, 110, 116, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        authSessionId = str(authSessionId).encode()
        sqrd += [11, 0, 1] + self.getIntBytes(len(authSessionId))
        for value2 in authSessionId:
            sqrd.append(value2)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def verifyEapLogin(self, authSessionId, type, accessToken):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/ACCT/authfactor/eap/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 14, 118, 101, 114, 105, 102, 121, 69, 97, 112, 76, 111, 103, 105, 110, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        authSessionId = str(authSessionId).encode()
        sqrd += [11, 0, 1] + self.getIntBytes(len(authSessionId))
        for value2 in authSessionId:
            sqrd.append(value2)
        sqrd += [12, 0, 2]
        sqrd += [8, 0, 1] + self.getIntBytes(type) # 1: FB  2: APPLE
        accessToken = str(accessToken).encode()
        sqrd += [11, 0, 2] + self.getIntBytes(len(accessToken))
        for value2 in accessToken:
            sqrd.append(value2)
        sqrd += [0, 0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def inviteFriends(self, friendMids, message, messageMetadata={}, imageObsPath="/r/myhome/c/0f3a02b6f993d3b627eeca97d2095b9b"):
        """ old ? """
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/PY3"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 13, 105, 110, 118, 105, 116, 101, 70, 114, 105, 101, 110, 100, 115, 0, 0, 0, 0]
        sqrd += [15, 0, 1, 11, 0, 0, 0, len(friendMids)]
        for mid in friendMids:
            sqrd += [0, 0, 0, len(mid)]
            for value in mid:
                sqrd.append(ord(value))
        sqrd += [11, 0, 2] + self.getStringBytes(message)
        _keys = messageMetadata.copy().keys()
        sqrd += [13, 0, 3, 11, 11] + self.getIntBytes(len(_keys))# key and val must str
        for _k in _keys:
            _v = messageMetadata[_k]
            sqrd += self.getStringBytes(_k)
            sqrd += self.getStringBytes(_v)
        sqrd += [11, 0, 4] + self.getStringBytes(imageObsPath)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        #data = self.decData(res.content)
        return self.tryReadData(data)
        
    def getCountrySettingV4(self):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/PY3"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getCountrySettingV4') + [0, 0, 0, 0]
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def getRSAKeyInfo(self, provider=1):
        """
        provider:
         - UNKNOWN(0),
         - LINE(1),
         - NAVER_KR(2),
         - LINE_PHONE(3)
        """
        _headers = {
            'x-lpqs': "/api/v3/TalkService.do"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getRSAKeyInfo') + [0, 0, 0, 0]
        sqrd += [8, 0, 2] + self.getIntBytes(provider)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
    
    def loginV2(self, keynm, encData, secret, deviceName='Chrome', cert=None, verifier=None):
        loginType = 2
        if verifier is not None:
            loginType = 1
        params = [
            [12, 2, [
                [8, 1, loginType],
                [8, 2, 1], # provider
                [11, 3, keynm],
                [11, 4, encData],
                [2, 5, 0],
                [11, 6, ''],
                [11, 7, deviceName],
                [11, 8, cert],
                [11, 9, verifier],
                [11, 10, secret],
                [8, 11, 1],
                [11, 12, "System Product Name"],
            ]]
        ]
        sqrd = self.generateDummyProtocol('loginV2', params, 3)
        return self.postPackDataAndGetUnpackRespData("/api/v3p/rs" ,sqrd, 3)
        
    def loginZ(self, keynm, encData, systemName='DeachSword-2021', certificate=None, verifier=None):
        _headers = {
            'x-lpqs': "/api/v3p/rs"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('loginZ') + [0, 0, 0, 0]
        sqrd += [12, 0, 2]
        loginType = 0
        if verifier is not None:
            loginType = 1
        sqrd += [8, 0, 1] + self.getIntBytes(loginType) # 2 if e2ee
        sqrd += [8, 0, 2] + self.getIntBytes(1) #provider
        sqrd += [11, 0, 3] + self.getStringBytes(keynm)
        sqrd += [11, 0, 4] + self.getStringBytes(encData)
        sqrd += [2, 0, 5, 0]
        sqrd += [11, 0, 6] + self.getStringBytes("") #accessLocation
        sqrd += [11, 0, 7] + self.getStringBytes(systemName)
        sqrd += [11, 0, 8] + self.getStringBytes(certificate)
        if verifier is not None:
            sqrd += [11, 0, 9] + self.getStringBytes(verifier)
        #sqrd += [11, 0, 10] + self.getStringBytes("") #secret
        sqrd += [8, 0, 11] + self.getIntBytes(1)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def checkLoginZPinCode(self, accessSession):
        _headers = {
            'X-Line-Access': accessSession,
            'x-lpqs': "/Q" #LF1
        }
        a = self.encHeaders(_headers)
        sqr_rd = a
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        hr = self.server.additionalHeaders(self.server.Headers, {
            'x-lhm': 'GET'
        })
        res = self.server.postContent(self.url, data=data, headers=hr)
        if res.status_code != 200:
            raise Exception("checkLoginZPinCode failed")
        data = self.decData(res.content)
        return json.loads(data[4:].split(b'\n', 1)[0].decode())['result']
        
    def checkLoginV2PinCode(self, accessSession):
        _headers = {
            'X-Line-Access': accessSession,
            'x-lpqs': "/LF1"
        }
        a = self.encHeaders(_headers)
        sqr_rd = a
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        hr = self.server.additionalHeaders(self.server.Headers, {
            'x-lhm': 'GET'
        })
        res = self.server.postContent(self.url, data=data, headers=hr)
        if res.status_code != 200:
            raise Exception("checkLoginV2PinCode failed")
        data = self.decData(res.content)
        return json.loads(data[4:].split(b'\n', 1)[0].decode())['result']
        
    def testFunc(self, path, funcName, funcValue=None, funcValueId=1):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': path
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, len(funcName)]
        for name in funcName:
            sqrd.append(ord(name))
        sqrd += [0, 0, 0, 0]
        print(sqrd)
        if funcValue:
            sqrd += [11, 0, funcValueId, 0, 0, 0, len(funcValue)]
            for value in funcValue: # string only
                sqrd.append(ord(value))
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
    
    def testTBinary(self):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/P3"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1, 0, 0, 0, 10]
        for value in "getProfile":
            sqrd.append(ord(value))
        sqrd += [0,0,0,0,0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        return data
    
    def testTCompact(self):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/P4"
        }
        a = self.encHeaders(_headers)
        sqrd = [130, 33, 1, 10]
        for value in "getProfile":
            sqrd.append(ord(value))
        #for value in "fetchOps":
        #    sqrd.append(ord(value))
        #sqrd += [38, 136, 176, 2, 21, 200, 1, 22, 238, 179, 106, 22, 226, 1, 0] fetchOps
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        self.tryReadTCompactData(data)
        return data
    
    def testTMoreCompact(self):
        _headers = {
            'X-Line-Access': self.authToken, 
            'x-lpqs': "/S5"
        }
        a = self.encHeaders(_headers)
        sqrd = [130, 33, 1] + self.getStringBytes('getProfile', True)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent(self.url, data=data, headers=self.server.Headers)
        data = self.decData(res.content)
        tmore = self.TMoreCompactProtocol(data)
        data = tmore.res
        return data


# -*- coding: utf-8 -*-
from datetime import datetime
import copy
import json, time, ntpath, re, hashlib
from base64 import b64encode

class Object(object):

    def __init__(self):
        pass
    
    """ 
    
    TimelineObs 
    Source: https://github.com/DeachSword/LINE-DemoS-Bot/blob/master/api/obs/obs.py
    
    """
    
    def updateProfileImage(self, path):
        url = self.LINE_OBS_DOMAIN + f'/r/talk/p/{self.profile[1]}'
        file = open(path, 'rb')
        hstr = 'DearSakura_%s' % int(time.time() * 1000)
        file_name = hashlib.md5(hstr.encode()).hexdigest()
        params = {
            'name': file_name,
            'quality': '100',
            "type": "image",
            "ver": "2.0"
        }
        r = self.server.postContent(url, headers=self.server.timelineHeaders, data={'params': json.dumps(params)}, files={'file':file})
        if r.status_code != 201:
            raise Exception(f"updateProfileImage failure. Receive statue code: {r.status_code}")
        return True
    
    def updateProfileCover(self, path):
        hstr = 'DearSakura_%s' % int(time.time() * 1000)
        objid = hashlib.md5(hstr.encode()).hexdigest()
        if not objid:
            hstr = 'DearSakura_%s' % int(time.time()*1000)
            objid = hashlib.md5(hstr.encode()).hexdigest()
        url = self.LINE_OBS_DOMAIN + f'/r/myhome/c/{objid}'
        file = {'file': open(path, 'rb')}
        params = {
            'name': objid,
            'quality': '100',
            'type': 'image',
            'ver': '2.0'
        }
        data = {'params': json.dumps(params)}
        r = self.server.postContent(url, headers=self.server.timelineHeaders, data=data, files=file)
        if r.status_code != 201:
            raise Exception(f"Upload object home failure. Receive statue code: {r.status_code}")
        objId = r.headers['x-obs-oid']
        home = self.updateProfileCoverById(objId)
        return home
    
    def updateImageToAlbum(self, mid, albumId, path):
        pass
        #privie
    
    def uploadObjHome(self, path, type='image', objId=None):
        if type not in ['image','video','audio']:
            raise Exception('Invalid type value')
        if type == 'image':
            contentType = 'image/jpeg'
        elif type == 'video':
            contentType = 'video/mp4'
        elif type == 'audio':
            contentType = 'audio/mp3'
        if not objId:
            hstr = 'DearSakura_%s' % int(time.time()*1000)
            objid = hashlib.md5(hstr.encode()).hexdigest()
        file = open(path, 'rb').read()
        params = {
            'name': '%s' % str(time.time()*1000),
            'userid': '%s' % self.profile[1],
            'oid': '%s' % str(objId),
            'type': type,
            'ver': '1.0' #2.0 :p
        }
        hr = self.server.additionalHeaders(self.server.timelineHeaders, {
            'Content-Type': contentType,
            'Content-Length': str(len(file)),
            'x-obs-params': self.genOBSParams(params,'b64') #base64 encode
        })
        r = self.server.postContent(self.LINE_OBS_DOMAIN + '/myhome/c/upload.nhn', headers=hr, data=file)
        if r.status_code != 201:
            raise Exception(f"Upload object home failure. Receive statue code: {r.status_code}")
        return objId
    
    def uploadMultipleImageToTalk(self, paths, to):
        if type(paths) != list:
            raise Exception('paths must be list')
        sqrd_base = [13, 0, 18, 11, 11]
        hashmap = {
            "GID" : "0",
            "GSEQ": "1",
            "GTOTAL": str(len(paths))
        }
        sqrd = sqrd_base + self.getIntBytes(len(hashmap.keys()))
        for hm in hashmap.keys():
            sqrd += self.getStringBytes(hm)
            sqrd += self.getStringBytes(hashmap[hm])
        sqrd += [0]
        data = bytes(sqrd)
        msg = json.dumps({
            "message": b64encode(data).decode('utf-8')
        })
        talkMeta = b64encode(msg.encode('utf-8')).decode('utf-8')
        res = self.uploadObjTalk(paths[0], to=to, talkMeta=talkMeta, returnHeaders=True)
        gid = res['x-line-message-gid']
        msgIds = [res['x-obs-oid']]
        if len(paths) > 1:
            hashmap["GID"] = gid
            nc = 2
            for img in paths[1:]:
                self.log(f"Upload image-{nc} with GID-{gid}...")
                hashmap["GSEQ"] = str(nc)
                sqrd = sqrd_base + self.getIntBytes(len(hashmap.keys()))
                for hm in hashmap.keys():
                    sqrd += self.getStringBytes(hm)
                    sqrd += self.getStringBytes(hashmap[hm])
                sqrd += [0]
                data = bytes(sqrd)
                msg = json.dumps({
                    "message": b64encode(data).decode('utf-8')
                })
                talkMeta = b64encode(msg.encode('utf-8')).decode('utf-8')
                res = self.uploadObjTalk(img, to=to, talkMeta=talkMeta, returnHeaders=True)
                msgIds.append(res['x-obs-oid'])
                nc += 1
        return gid
        
    def uploadObjTalk(self, path=None, type='image', objId=None, to=None, talkMeta=None, returnHeaders=False):
        if type not in ['image','gif','video','audio','file']:
            raise Exception('Invalid type value')
        headers=self.server.timelineHeaders
        files = {'file': open(path, 'rb')}
        #url = self.LINE_OBS_DOMAIN + '/talk/m/upload.nhn' #if reqseq not working
        url = self.LINE_OBS_DOMAIN + '/r/talk/m/reqseq'
        params = {
            "type": "image",
            "ver": "2.0",
            "name": files['file'].name,
            "oid": "reqseq",
            "reqseq": str(self.getCurrReqId()),
            "cat": "original"
        }
        if objId != None:
            params['oid'] = objId
        if to != None:
            params['tomid'] = to
        if type != 'gif':
            params['type'] = type
            data = {'params': self.genOBSParams(params)}
        elif type == 'gif':
            params = {
                'type': 'image',
                'ver': '2.0',
                'name': files['file'].name,
                'oid': 'reqseq',
                'reqseq': '%s' % str(self.getCurrReqId()),
                'tomid': '%s' % str(to),
                'cat': 'original'
            }
            files = None
            data = open(path, 'rb').read()
            headers = self.server.additionalHeaders(self.server.Headers, {
                'content-type': 'image/gif',
                'Content-Length': str(len(data)),
                'x-obs-params': self.genOBSParams(params,'b64'), #base64 encode
                'X-Line-Access': self.acquireEncryptedAccessToken()[7:]
            })
        if talkMeta != None:
            headers = self.server.additionalHeaders(headers, {
                'X-Talk-Meta': talkMeta
            })
        r = self.server.postContent(url, data=data, headers=headers, files=files)
        if r.status_code != 201:
            print("uploadObjTalk: ", r.status_code)
            raise Exception('Upload %s failure.' % type)
        else:
            if returnHeaders:
                return r.headers
            if objId is None:
                objId = r.headers['x-obs-oid'] #the message seq, if u oid using reqseq
            objHash = r.headers['x-obs-hash']  #for view on cdn
        return objId
    
    def forwardObjectMsg(self, to, msgId, contentType='image', objFrom='c'):
        if contentType not in ['image','video','audio','file']:
            raise Exception('Type not valid.')
        data = {
            'name': f'CHRLINE-{int(time.time())}', 'tomid': to,'oid': 'reqseq','reqseq': self.revision,'type': contentType,'copyFrom': f'/{"g2" if self.getToType(to) == 4 else "talk"}'
        }
        r = self.server.postContent(self.LINE_OBS_DOMAIN + '/g2/m/copy.nhn' if self.getToType(to) == 4 else '/talk/m/copy.nhn', data=data, headers=self.server.timelineHeaders)
        # self.LINE_HOST_DOMAIN + '/oa/r/talk/m/reqseq/copy.nhn'
        if r.status_code != 200:
            raise Exception(f'Forward object failure: {r.status_code}')
        return True
    
    def forwardKeepObjectMsg(self, to, oid, contentId, contentType='image'):
        if contentType not in ['image','video','audio','file']:
            raise Exception('Type not valid.')
        #data = copyFrom=/keep/p/linekeep_1317462115813022667381t0d0d34a0&tomid=c7bb0199d949fee69dd445eaad321cd2b&reqseq=14884&type=video&duration=3320&contentId=WQDnZP2&size=1606120
        data = {
            'copyFrom': f'/keep/p/{oid}', 
            'tomid': to,
            'reqseq': self.revision,
            'type': contentType,
            'contentId': contentId,
            'size': '-1',
        }
        if contentType == 'image':
            data['cat'] = 'original'
        elif contentType in ['audio', 'video']:
            data['duration'] = '3000'
        r = self.server.postContent(self.LINE_HOST_DOMAIN + '/oa/r/talk/m/reqseq/copy.nhn', data=data, headers=self.server.timelineHeaders)
        if r.status_code != 200:
            raise Exception(f'Forward object failure: {r.status_code}')
        return r.text

    def trainingImage(self, msgId):
        data = {
           "useAsTrainingData" : False,
           "input" : 'obs://talk/m/%s' % msgId,
           "filters" : {
              "OCR" : {
                 "params" : {
                    "useColorPicker" : "true"
                 }
              }
           }
        }
        hr = self.server.additionalHeaders(self.server.timelineHeaders, {
            "x-client-channel": "chat_viewer",
            "x-lal": "zh-Hant_TW",
            "x-recognition-input-type": "obs",
            "Content-Type": "application/json",
        })
        r = self.server.postContent(self.LINE_OBS_DOMAIN + '/px/talk/m/%s/recognition.obs' % msgId, data=json.dumps(data), headers=hr)
        if r.status_code != 200:
            raise Exception(f'Training image failure: {r.status_code}')
        return r.json()
        
    def downloadObjectMsg(self, objId, path, objFrom='c'):
        obs_path = f'/r/{"g2" if self.getToType(objFrom) == 4 else "talk"}/m/{objId}'
        r = self.server.getContent(self.LINE_OBS_DOMAIN + obs_path, headers=self.server.timelineHeaders)
        with open(path, 'wb') as f:
            f.write(r.content)
        return r.content




from Crypto.PublicKey import RSA
from Crypto.Cipher import AES
import Crypto.Cipher.PKCS1_OAEP as rsaenc
from base64 import b64encode, b64decode
from Crypto.Util.Padding import pad, unpad
from hashlib import md5, sha1
import xxhash
from datetime import datetime
import struct
import time
import json
import os
import rsa
import os
import binascii
import urllib
import base64
import axolotl_curve25519 as curve

import gevent.monkey
gevent.monkey.patch_all()
import requests # patch after

class Models(object):

    def __init__(self):
        self.lcsStart = "0005"
        self.le = "18"
        self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0LRokSkGDo8G5ObFfyKiIdPAU5iOpj+UT+A3AcDxLuePyDt8IVp9HpOsJlf8uVk3Wr9fs+8y7cnF3WiY6Ro526hy3fbWR4HiD0FaIRCOTbgRlsoGNC2rthp2uxYad5up78krSDXNKBab8t1PteCmOq84TpDCRmainaZQN9QxzaSvYWUICVv27Kk97y2j3LS3H64NCqjS88XacAieivELfMr6rT2GutRshKeNSZOUR3YROV4THa77USBQwRI7ZZTe6GUFazpocTN58QY8jFYODzfhdyoiym6rXJNNnUKatiSC/hmzdpX8/h4Y98KaGAZaatLAgPMRCe582q4JwHg7rwIDAQAB\n-----END PUBLIC KEY-----"
        self.key = RSA.importKey(self.PUBLIC_KEY)
        self.encryptKey = b"DearSakura+2021/"
        self.IV = bytes([78, 9, 72, 62, 56, 245, 255, 114, 128, 18, 123, 158, 251, 92, 45, 51])
        self.cipher = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV)
        self.d_cipher = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV)
        self.encEncKey()
        #self.initWithBiz()
        #self.initWithAndroid()
        
    def log(self, text, debugOnly=False):
        if debugOnly and not self.isDebug:
            return
        print(f"[{datetime.now():%Y-%m-%d %H:%M:%S}] {text}")
        
    def genOBSParams(self, newList, returnAs='json', ext='jpg'):
        oldList = {'name': f'CHRLINE-{int(time.time())}.{ext}','ver': '1.0'}
        if returnAs not in ['json','b64','default']:
            raise Exception('Invalid parameter returnAs')
        oldList.update(newList)
        if 'range' in oldList:
            new_range = 'bytes 0-%s\/%s' % ( str(oldList['range']-1), str(oldList['range']) )
            oldList.update({'range': new_range})
        if returnAs == 'json':
            oldList = json.dumps(oldList)
            return oldList
        elif returnAs == 'b64':
            oldList = json.dumps(oldList)
            return b64encode(oldList.encode('utf-8'))
        elif returnAs == 'default':
            return oldList
        
    def checkNextToken(self):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.tokens')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.authToken.encode()).hexdigest()
        if os.path.exists(savePath + f"/{fn}"):
            self.authToken = open(savePath + f"/{fn}", "r").read()
            self.log(f"New Token: {self.authToken}")
            self.checkNextToken()
        return self.authToken

    def handleNextToken(self, newToken):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.tokens')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.authToken.encode()).hexdigest()
        open(savePath + f"/{fn}", "w").write(newToken)
        self.authToken = newToken
        self.log(f"New Token: {newToken}")
        self.server.timelineHeaders['X-Line-Access'] = self.authToken
        self.server.timelineHeaders['X-Line-ChannelToken'] = self.issueChannelToken()[5] #need?
        
    def getCustomData(self):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.customDataId.encode()).hexdigest()
        if os.path.exists(savePath + f"/{fn}"):
            self.custom_data = json.loads(open(savePath + f"/{fn}", "r").read())
        self.log(f'Loading Custom Data: {fn}')
        return True
        
    def saveCustomData(self):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.customDataId.encode()).hexdigest()
        open(savePath + f"/{fn}", "w").write(json.dumps(self.custom_data))
        return True
        
    def getSqrCert(self):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = "cert.pem"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None
        
    def saveSqrCert(self, cert):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = "cert.pem"
        open(savePath + f"/{fn}", "w").write(cert)
        return True
        
    def getEmailCert(self, email):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{email}.crt"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None
        
    def saveEmailCert(self, email, cert):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{email}.crt"
        open(savePath + f"/{fn}", "w").write(cert)
        return True
    
    def initWithAndroid(self, ver=7):
        if ver == 1:
            self.lcsStart = "0001"
            self.le = "1"
            self.PUBLIC_KEY = '-----BEGIN PUBLIC KEY-----\nMIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCZAAoZNRwIlLUXaUgrgYi8bAYq\nQeFVtXvCNIEm+F4/jAyTU3YoDwmoLaKQ6itGOonykGtwy2k/3BeWefL/q5eUGjVG\nBEa1vBeUNEb4IFU8n9WK3N/GIIPuD6ZiusB+U1FPg/NaEiVX8ldmEQJgmuG1hykk\n2dU3oy7O1M+Kwl1lJQIDAQAB\n-----END PUBLIC KEY-----'
        elif ver == 4:
            self.lcsStart = "0004"
            self.le = "3" # LegyEncHelper.cpp::decryptStream -> legy xle value
            self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwpAwTVluR1Z++tVzxtOD\nr7XxSv6oqrwvj/8c8SkfFsS8zM7CvIT8j+x+6Qs1JjNRDtYjAwPKO3tO+qOAdA+8\n7FHpx0THDJIi4VYxSZ2uDh0U8Luxh02whwM8gPbPQNN3sEd5++kJ3cCh5eeAIiUd\nDrwPhHzxO8swpBRdxJB/pzibEqpG2U2764JlPscN9D896qmBN6CBRKpXk/MmUDAI\n4xg+uQk/ykn3SNXJSgQwI1UD9KuiR+X9tbJlKRMN5JpUrSuEwRPQQDMaWpSIdCJM\noFqJLNwt9b1RR/JEB01Eup+3QCub20/CObCmHZY6G26KTDHLoTRZ1xzymdYhdJ43\nCwIDAQAB\n-----END PUBLIC KEY-----"
        elif ver == 6:
            self.lcsStart = "0007"
            self.le = "7"
            self.PUBLIC_KEY = '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA5ABzJbexh+HH1+RzVTH4\nDFj8b/42vRqUp0NWLIBAgi5+bAgeJYzyVBI7Pk6YkQnd44HPvUvFMF3V3TocRXLP\nZV/NSawgcAh+VrXe3VIlruCOe14qrd/ZMpTRTxtBJ5aRpVhTsnGpZtGggPYnPh4c\n6V/R7Wxymj4SBZ1Ipsa7ZI83z/XIPFXT38qTJN3UAW5YhjQon1eNbwaxALVajuvI\npUE52aIBi05gE/V0HEoOUjfOg1V6RHFbcchTgmcze6Vbye+7kmdsIboDXnNpm/fJ\nuItub+iwLKSWf9OPc/bYpU4YVBxZXzmSCXFMLeCe2i5wJeMg4iG8NpVpwVv2W+Hb\nQQIDAQAB\n-----END PUBLIC KEY-----'
        elif ver == 7:
            self.lcsStart = "0008"
            self.le = "7"
            self.PUBLIC_KEY = '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsMC6HAYeMq4R59e2yRw6\nW1OWT2t9aepiAp4fbSCXzRj7A29BOAFAvKlzAub4oxN13Nt8dbcB+ICAufyDnN5N\nd3+vXgDxEXZ/sx2/wuFbC3B3evSNKR4hKcs80suRs8aL6EeWi+bAU2oYIc78Bbqh\nNzx0WCzZSJbMBFw1VlsU/HQ/XdiUufopl5QSa0S246XXmwJmmXRO0v7bNvrxaNV0\ncbviGkOvTlBt1+RerIFHMTw3SwLDnCOolTz3CuE5V2OrPZCmC0nlmPRzwUfxoxxs\n/6qFdpZNoORH/s5mQenSyqPkmH8TBOlHJWPH3eN1k6aZIlK5S54mcUb/oNRRq9wD\n1wIDAQAB\n-----END PUBLIC KEY-----'
        self.key = RSA.importKey(self.PUBLIC_KEY)
        self.IV = bytes([78, 9, 72, 62, 56, 245, 255, 114, 128, 18, 123, 158, 251, 92, 45, 51])
        self.encryptKey = b"DearSakura+2021/"
        self.encEncKey()
        print(self._encryptKey)

    def encHeaders(self, headers):
        t = headers.keys()
        data = []
        self.mFhrnmxnNF(len(t), data)
        for i in t:
            self.mFhrnmxnNF(len(i), data)
            self.wYEpEYldst(i, data)
            self.mFhrnmxnNF(len(headers[i]), data)
            self.wYEpEYldst(headers[i], data)
        o = len(data);
        data = [255 & o] + data
        data = [255 & o >> 8] + data
        return data

    def decHeaders(self, data):
        headers = {}
        tbin = self.TBinaryProtocol()
        tbin.data = data
        dataLen = tbin.readI16() + 2
        headerLen = tbin.readI16()
        for i in range(headerLen):
            size = tbin.readI16()
            _k = tbin.y(size)
            size = tbin.readI16()
            _v = tbin.y(size)
            headers[_k] = _v
        return headers, data[dataLen:]

    def encEncKey(self):
        # heh
        a = rsaenc.new(self.key)
        self._encryptKey = self.lcsStart + b64encode(a.encrypt(self.encryptKey)).decode()
        
    def encData(self, data):
        _data = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV).encrypt(pad(data, AES.block_size))
        debug = []
        return _data + self.XQqwlHlXKK(self.encryptKey, _data)
        
    def decData(self, data):
        data = pad(data, AES.block_size)
        _data = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV).decrypt(data)[:-16]
        _data = unpad(_data, AES.block_size)
        i = 1
        data = self.yVdzCLDwMN(_data, i)
        i = 3
        return _data
        
    def mFhrnmxnNF(self, t, e):
        i = 65536 
        if t < -1 * 32768 or t >= i:
            raise Exception(t + " is incorrect for i16.")
        e.append(255 & t >> 8)
        e.append(255 & t)
        
    def wYEpEYldst(self, t, e):
        for i in range(len(t)):
            e.append(ord(t[i]))
        
    def xZVpUuXFru(t):
        if 8 == len(t):
            return t
        e = ""
        i = 0
        n = 8 - len(t)
        while i < n:
            e += "0"
            i += 1
        return e + t
        
    def pmAWhahfKx(self, t):
        e = []
        i = 0 
        n = len(t)
        while i < n:
            _i = 0
            try:
                _i = int(t[i:i + 2], 16)
            except:
                _i = 16
            e.append(_i);
            i += 2
        return e
        
    def XQqwlHlXKK(self, e, i):
        r = []
        for o in range(16):
            r.append(92 ^ e[o])
        n = xxhash.xxh32(b'',seed=0)
        s = xxhash.xxh32(b'',seed=0)
        n.update(bytes(r))
        for o in range(16):
            r[o] ^= 106
        s.update(bytes(r))
        s.update(i)
        a = s.hexdigest() # is b8a7c677?
        n.update(bytes(self.pmAWhahfKx(a)))
        c = n.hexdigest() # is 3f97d2f6?
        d = self.pmAWhahfKx(c)
        return bytes(d)
        
    def yVdzCLDwMN(self, d, i):
        return (255 & self.xnEmbaRWhy(d, i)) << 8 | 255 & self.xnEmbaRWhy(d, i+1)
    
        
    def xnEmbaRWhy(self, d, i):
        t = d[i];
        if t > 127:
            t = 0 - (t - 1 ^ 255)
        return t
    
    def generateDummyProtocol(self, name, params, type):
        if type == 3:
            data = [128, 1, 0, 1] + self.getStringBytes(name) + [0, 0, 0, 0]
        elif type == 4:
            data = [130, 33, 00] + self.getStringBytes(name, isCompact=True)
        data += self.generateDummyProtocolField(params, type) + [0]
        return data
    
    def generateDummyProtocolField(self, params, type):
        isCompact = False
        data = []
        tcp = self.TCompactProtocol()
        for param in params:
            # [10, 2, revision]
            _type = param[0]
            _id = param[1]
            _data = param[2]
            # if _data is None:
                # continue
            if type == 3:
                data += [_type, 0, _id]
                isCompact = False
            elif type == 4:
                data += tcp.getFieldHeader(tcp.CTYPES[_type], _id)
                isCompact = True
            data += self.generateDummyProtocolData(_data, _type, isCompact)
        return data
    
    def generateDummyProtocolData(self, _data, type, isCompact=False):
        data = []
        tcp = self.TCompactProtocol()
        ttype = 4 if isCompact else 3
        if type == 2:
            if isCompact:
                _compact = self.TCompactProtocol()
                a = _compact.getFieldHeader(1 if _data == True else 2, 0)
            else:
                data += [1] if _data == True else [0]
        elif type == 8:
            data += self.getIntBytes(_data, isCompact=isCompact)
        elif type == 10:
            data += self.getIntBytes(_data, 8, isCompact=isCompact)
        elif type == 11:
            data += self.getStringBytes(_data, isCompact=isCompact)
        elif type == 12:
            data += self.generateDummyProtocolField(_data, ttype) + [0]
        elif type == 13:
            _ktype = _data[0]
            _vtype = _data[1]
            _vdata = _data[2]
            if isCompact:
                data += tcp.writeMapBegin(_ktype, _vtype, len(_vdata))
            else:
                data += [_ktype, _vtype] + self.getIntBytes(len(_vdata), isCompact=isCompact)
            for vd in _vdata:
                data += self.generateDummyProtocolData(vd, _ktype, isCompact)
                data += self.generateDummyProtocolData(_vdata[vd], _vtype, isCompact)
        elif type == 14 or type == 15:
            # [11, targetUserMids]
            _vtype = _data[0]
            _vdata = _data[1]
            if isCompact:
                data += tcp.writeCollectionBegin(_vtype, len(_vdata))
            else:
                data += [_vtype] + self.getIntBytes(len(_vdata), isCompact=isCompact)
            for vd in _vdata:
                data += self.generateDummyProtocolData(vd, _vtype, isCompact)
        else:
            raise Exception(f"[generateDummyProtocolData] not support type: {type}")
        return data
        
    def postPackDataAndGetUnpackRespData(self, path: str, bdata: bytes, ttype: int = 3, encType=None, headers=None, access_token=None):
        self.log(f"--> POST {path}", True)
        if headers is None:
            headers = self.server.Headers.copy()
        if access_token is None:
            access_token = self.authToken
        ptype = "TBINARY" if ttype == 3 else "TCOMPACT"
        headers["content-type"] = "application/x-thrift; protocol=" + ptype
        if encType is None:
            encType = self.encType
        if encType == 0:
            data = bytes(bdata)
            if "x-le" in headers:
                del headers['x-le']
                del headers['x-lcs']
            if access_token is not None:
                headers['X-Line-Access'] = access_token
            res = self.req_h2.post(self.LINE_GW_HOST_DOMAIN + path, data=data, headers=headers, timeout=180)
            data = res.content
        elif encType == 1:
            if access_token is not None:
                _headers = {
                    'X-Line-Access': access_token, 
                    'x-lpqs': path
                }
            else:
                _headers = {
                    'x-lpqs': path
                }
            a = self.encHeaders(_headers)
            b = bdata
            c = a + b
            _data = bytes(c)
            data = self.encData(_data)
            headers['x-cl'] = str(len(data))
            res = self.req.post(self.LINE_GF_HOST_DOMAIN + self.LINE_ENCRYPTION_ENDPOINT, data=data, headers=headers)
            data = self.decData(res.content)
        else:
            raise Exception(f"Unknown encType: {encType}")
        self.log(f"<--  {res.status_code}", True)
        self.log(f"{data}", True)
        if res.status_code == 200:
            if res.headers['x-lc'] not in ['200', '410']:
                raise Exception(f'Invalid response code: {res.headers["x-lc"]}')
            if encType == 1:
                respHeaders, data = self.decHeaders(data)
            else:
                respHeaders = res.headers
            if 'x-line-next-access' in respHeaders:
                self.handleNextToken(respHeaders['x-line-next-access'])
            res = None
            if ttype == 3:
                res = self.TBinaryProtocol(data).res
            elif ttype == 4:
                res = self.tryReadTCompactData(data)
            elif ttype == 5:
                res = self.TMoreCompactProtocol(data).res
            else:
                raise Exception(f"Unknown ThriftType: {ttype}")
            if type(res) == dict and 'error' in res:
                print(res['error'])
                if res['error']['message'] is not None and (res['error']['message'] in ["EXPIRED", "REVOKE", "LOG_OUT", "AUTHENTICATION_DIVESTED_BY_OTHER_DEVICE", "DEVICE_LOSE", "IDENTIFY_MODIFIED", "V3_TOKEN_CLIENT_LOGGED_OUT", "DELETED_ACCOUNT"] or res['error']['message'].startswith('suspended:')):
                    self.is_login = False
                    self.log(f"LOGIN OUT: {res['error']['message']}")
                    raise Exception(res['error'])
                else:
                    print(res['error']['message'])
            return res
        else:
            print(data)
            return None
        
    def getCurrReqId(self):
        self._msgSeq = 0
        if "_reqseq" in self.custom_data:
            self._msgSeq = self.custom_data["_reqseq"]
        self._msgSeq += 1
        self.custom_data["_reqseq"] = self._msgSeq
        self.saveCustomData()
        return self._msgSeq
        
    def getIntBytes(self, i, l=4, isCompact=False):
        i = int(i)
        if isCompact:
            _compact = self.TCompactProtocol()
            a = _compact.makeZigZag(i, 32 if l**2 == 16 else 64)
            b = _compact.writeVarint(a)
            return b
        _seq = int(i).to_bytes(l, byteorder="big")
        res = []
        for value in _seq:
            res.append(value)
        return res
        
    def getStringBytes(self, text, isCompact=False):
        if text is None:
            text = ""
        if type(text) == bytes:
            pass
        else:
            text = str(text).encode()
        if isCompact:
            _compact = self.TCompactProtocol()
            sqrd = _compact.writeVarint(len(text))
        else:
            sqrd = self.getIntBytes(len(text))
        for value in text:
            sqrd.append(value)
        return sqrd
        
    def getFloatBytes(self, val):
        res = []
        for value in struct.pack('!d', val):
            res.append(value)
        return res
        
    def getMagicStringBytes(self, val, rev=False):
        res = []
        i = 0
        if rev:
            res = binascii.b2a_hex(val)
        else:
            if len(val) == 32:
                for ii in range(16):
                    iii = ii * 2
                    i = iii + 1
                    mgc = (int(val[iii], 16) << 4) + int(val[i], 16)
                    res.append(mgc)
            else:
                raise Exception(f"getMagicStringBytes() expected 32, but got {len(val)}")
        return res

    def createSqrSecret(self, base64Only=False):
        private_key = curve.generatePrivateKey(os.urandom(32))
        public_key = curve.generatePublicKey(private_key)
        secret = urllib.parse.quote(b64encode(public_key).decode())
        version = 1
        if base64Only:
            return [private_key, b64encode(public_key).decode()]
        return [private_key, f"?secret={secret}&e2eeVersion={version}"]

    def getE2EESelfKeyData(self, mid):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{mid}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        keys = self.getE2EEPublicKeys()
        for key in keys:
            _keyData = self.getE2EESelfKeyDataByKeyId(key[2])
            if _keyData is not None:
                return _keyData
        return None

    def getE2EESelfKeyDataByKeyId(self, keyId):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"key_{keyId}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        return None

    def saveE2EESelfKeyData(self, mid, pubK, privK, kI, e2eeVersion):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{mid}.json"
        fn2 = f"key_{kI}.json"
        data = json.dumps({
            "keyId": kI,
            "privKey": b64encode(privK).decode(),
            "pubKey": b64encode(pubK).decode(),
            "e2eeVersion": e2eeVersion,
        })
        if mid is not None:
            open(savePath + f"/{fn}", "w").write(data)
        open(savePath + f"/{fn2}", "w").write(data)
        return True

    def getCacheData(self, cT, cN, needHash=True):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), cT)
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{cN}"
        if needHash:
            fn = md5(cN.encode()).hexdigest()
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveCacheData(self, cT, cN, cD, needHash=True):
        savePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), cT)
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{cN}"
        if needHash:
            fn = md5(cN.encode()).hexdigest()
        data = cD
        open(savePath + f"/{fn}", "w").write(data)
        return True
    
    def decodeE2EEKeyV1(self, data, secret, mid=None):
        if 'encryptedKeyChain' in data:
            encryptedKeyChain = base64.b64decode(data['encryptedKeyChain'])
            #hashKeyChain = data['hashKeyChain']
            keyId = data['keyId']
            publicKey = base64.b64decode(data['publicKey'])
            e2eeVersion = data['e2eeVersion']
            e2eeKey = self.decryptKeyChain(publicKey, secret, encryptedKeyChain)
            print(f"E2EE Priv Key: {e2eeKey[0]}")
            print(f"E2EE Pub Key: {e2eeKey[1]}")
            print(f"keyId: {keyId}")
            print(f"e2eeVersion: {e2eeVersion}")
            self.saveE2EESelfKeyData(mid, e2eeKey[1], e2eeKey[0], keyId, e2eeVersion)
            return {
                "keyId": keyId,
                "privKey": e2eeKey[0],
                "pubKey": e2eeKey[1],
                "e2eeVersion": e2eeVersion,
            }
        
    def tryReadData(self, data, mode=1):
        _data = {}
        if mode == 0:
            data = bytes(4) + data + bytes(4)
        if data[4] == 128:
            a = 12 + data[11]
            b = data[12:a].decode()
            _data[b] = {}
            c = data[a + 4]
            if c == 0:
                return None
            id = data[a + 6]
            if id == 0:
                if c == 10:
                    a = int.from_bytes(data[a + 9:a + 15], "big")
                    _data[b] = a
                elif c == 11:
                    d = data[a + 10]
                    e = data[a + 11:a + 11 + d].decode()
                    _data[b] = e
                elif c == 12:
                    _data[b] = self.readContainerStruct(data[a + 7:])
                elif c == 13:
                    _data[b] = self.readContainerStruct(data[a + 4:])
                elif c == 14 or c == 15:
                    _data[b] = self.readContainerStruct(data[a + 4:], stopWithFirst=True)[0]
                else:
                    print(f"[tryReadData]不支援Type: {c} => ID: {id}")
            else:
                if c != 0:
                    error = {}
                    if c == 11:
                        t_l = data[a + 10]
                        error = data[a + 11:a + 11 + t_l].decode()
                    else:
                        ed = self.readContainerStruct(data[a + 4:])[1]
                        error = {
                            'code': ed.get(1),
                            'message': ed.get(2),
                            'metadata': ed.get(3),
                            '_data': ed
                        }
                    _data[b] = {
                        "error": error
                    }
            return _data[b]
        else:
            if data[6:24] == b"x-line-next-access":
                a = data[25]
                b = data[26:26 + a]
                self.handleNextToken(b.decode())
                data = bytes([0, 0, 0, 0]) + data[26 + a:]
                return self.tryReadData(data)
        return _data
        
    def readContainerStruct(self, data, get_data_len=False, stopWithFirst=False):
        _data = {}
        nextPos = 0
        if len(data) < 3:
            return None
        dataType = data[0]
        id = data[2]
        #print(f"{id} -> {dataType}")
        if data[0] == 2:
            a = data[3]
            if a == 1:
                _data[id] = True
            else:
                _data[id] = False
            nextPos = 4
        elif data[0] == 3:
            a = int.from_bytes(data[3:4], "big")
            _data[id] = a
            nextPos = 4
        elif data[0] == 4:
            a = data[3:11]
            a = struct.unpack('!d', a)[0]
            _data[id] = a
            nextPos = 11
        elif data[0] == 8:
            a, = struct.unpack('!i', data[3:7])
            _data[id] = a
            nextPos = 7
        elif data[0] == 10:
            a, = struct.unpack('!q', data[3:11])
            _data[id] = a
            nextPos = 11
        elif data[0] == 11:
            a = int.from_bytes(data[5:7], "big")
            if a == 0:
                _data[id] = ''
                nextPos = a + 7
            else:
                b = data[7:a+7]
                try:
                    _data[id] = b.decode()
                except:
                    _data[id] = b
                nextPos = a + 7
        elif data[0] == 12:
            if data[3] == 0:
                _data[id] = {}
                nextPos = 4
            else:
                a = self.readContainerStruct(data[3:], True)
                _data[id] = a[0]
                nextPos = a[1] + 4
        elif data[0] == 13:
            # dict
            # 0D 00 24 0B 0B 00 00 00 02 00 00 00 07
            kt = data[3] # key type
            a = data[4] # value type
            b, = struct.unpack('!i', data[5:9]) # count
            c = 9
            _d = {}
            if b != 0:
                #print(f"ktype: {kt}")
                #print(f"kvalue: {a}")
                for d in range(b):
                    if True:
                        __key = self.readContainerStruct(bytes([kt, 0, 0]) + data[c:], get_data_len=True, stopWithFirst=True)
                        _key = __key[0][0]
                        vp = c + __key[1] - 3 # value pos
                        __value = self.readContainerStruct(bytes([a, 0, 0]) + data[vp:], get_data_len=True, stopWithFirst=True)
                        _value = __value[0][0]
                        c = vp + __value[1] - 3
                    # old code...
                    elif kt == 8:
                        # f = c + 1
                        # g = data[c + 4]
                        _key = data[c]
                        _value = self.readContainerStruct(bytes([a, 0, 0]) + data[f + 1:])[0]
                        # h = f + 4 + g
                        # _value = data[f + 4:h].decode()
                        c += 5
                    else:
                        g = int.from_bytes(data[f + 1:f + 5], "big") # value len
                        _key = data[c + 1:f + 1].decode()
                        h = f + g + 5
                        if a == 10:
                            __value = int.from_bytes(data[f+1:f+9], "big")
                            _value = __value
                            h = f + 9
                            c = h + 3
                        elif a == 12:
                            __value = self.readContainerStruct(data[f+1:], True)
                            _value = __value[0]
                            h = f + __value[1]
                            c = h
                        elif a == 15:
                            __value = self.readContainerStruct(data[f+1:], True)
                            _value = __value[0]
                            h = f + __value[1]
                            c = h + 1
                        else:
                            _value = data[f + 5:h].decode()
                            c = h + 3
                    _d[_key] = _value
                _data[id] = _d
                nextPos = c
                # old code...
                # if a in [10, 11]:
                #     nextPos -= 3
            else:
                nextPos = 9
                _data[id] = {}
        elif data[0] == 14:
            type = data[3]
            count = int.from_bytes(data[4:8], "big")
            _data[id] = []
            nextPos = 8
            if count != 0:
                for i in range(count):
                    if type == 8:
                        a = 0
                        b = self.readContainerStruct(bytes([type, 0, 0]) + data[nextPos:])[0]
                    else:
                        a = int.from_bytes(data[nextPos:nextPos + 4], "big")
                        b = data[nextPos + 4:nextPos + 4 + a].decode()
                    _data[id].append(b)
                    nextPos += 4 + a
        elif data[0] == 15:
            type = data[3]
            d = data[7]
            _data[id] = []
            e = 8
            for _d in range(d):
                if type == 8:
                    f = int.from_bytes(data[e:e+4], "big")
                    _data[id].append(f)
                    e += 4
                elif type == 11:
                    f = data[e+3]
                    dd = data[e+4:e+4+f]
                    try:
                        dd = dd.decode()
                    except:
                        pass
                    _data[id].append(dd)
                    e += f + 4
                elif type == 12:
                    f = self.readContainerStruct(data[e:], True)
                    _data[id].append(f[0])
                    if f[2] in [12, 13]:
                        e += f[1] + 1
                    else:
                        e += f[1] + 1
                else:
                    print(f"[readContainerStruct_LIST(15)]不支援Type: {type}")
            nextPos += e
        elif data[0] != 0:
            print(f"[readContainerStruct]不支援Type: {data[0]} => ID: {id}")
        if nextPos > 0 and not stopWithFirst:
            data = data[nextPos:]
            if len(data) > 2:
                c = self.readContainerStruct(data, True)
                if c[0]:
                    _data.update(c[0])
                    nextPos += c[1] # lol, why i forget it
                    if c[2] != 0:
                        dataType = c[2]
        if get_data_len:
            return [_data, nextPos, dataType]
        return _data
        
    def tryReadTCompactData(self, data):
        _data = {}
        data = bytes(4) + data
        if data[4] == 130:
            a = 8 + data[7]
            b = data[8:a].decode()
            _data[b] = {}
            _dec = self.TCompactProtocol()
            (fname, ftype, fid, offset) = _dec.readFieldBegin(data[a:])
            offset += a + 1
            if ftype == 12:
                _data[b] = self.tryReadTCompactContainerStruct(data[a:])
                if 0 in _data[b]:
                    _data[b] = _data[b][0]
                else:
                    error = {
                        'code': _data[b][1][1],
                        'message': _data[b][1].get(2, None),
                        'metadata': _data[b][1].get(3, None)
                    }
                    _data[b] = {
                        "error": error
                    }
            return _data[b]
        return None
        
    def tryReadTCompactContainerStruct(self, data, id=0, get_data_len=False):
        _data = {}
        _dec = self.TCompactProtocol()
        (fname, ftype, fid, offset) = _dec.readFieldBegin(data)
        nextPos = 0
        fid += id
        if ftype == 0:
            pass
        elif ftype == 1:
            _data[fid] = _dec.readBool()
            nextPos = 1
        elif ftype == 2:
            _data[fid] = _dec.readBool()
            nextPos = 1
        elif ftype == 5:
            (_data[fid], nextPos) = _dec.readI32(data[offset:], True)
            nextPos += 1
        elif ftype == 6:
            (_data[fid], nextPos) = _dec.readI64(data[offset:], True)
            #nextPos += -2
        elif ftype == 8:
            (_data[fid], nextPos) = _dec.readBinary(data[offset:])
        elif ftype == 9 or ftype == 10:
            ### todo:
            #       ftype == 10 == SET
            (vtype, vsize, vlen) = _dec.readCollectionBegin(data[offset:])
            offset += vlen
            _data[fid] = []
            _nextPos = 0
            for i in range(vsize):
                if vtype == 8:
                    (__data, _nextPos) = _dec.readBinary(data[offset:])
                    _data[fid].append(__data)
                    offset += _nextPos - 1
            nextPos += offset
        elif ftype == 12:
            (__data, nextPos) = self.tryReadTCompactContainerStruct(data[offset:], get_data_len=True)
            nextPos += 2
            _data[fid] = __data
        elif ftype != 0:
            print(f"[tryReadTCompactContainerStruct]不支援Type: {ftype} => ID: {fid}")
        if nextPos > 0:
            data = data[nextPos:]
            c = self.tryReadTCompactContainerStruct(data, id=fid, get_data_len=True)
            if c[0]:
                _data.update(c[0])
                nextPos += c[1]
        if get_data_len:
            return [_data, nextPos]
        return _data
        
    def tryReadThriftContainerStruct(self, data, id=0, get_data_len=False):
        _data = {}
        _dec = self.TCompactProtocol()
        ftype = data[0] & 15
        fid = (data[0] >> 4) + id
        offset = 1
        nextPos = 0
        if ftype == 0:
            _data = None
        elif ftype == 4:
            _data[fid] = _dec.readDouble(data[offset:])
            nextPos += 8
        elif ftype == 5:
            (_data[fid], nextPos) = _dec.readI32(data[offset:], True)
            nextPos += 1
        elif ftype == 6:
            (_data[fid], nextPos) = _dec.readI64(data[offset:], True)
            nextPos += 1
        elif ftype == 8:
            (_data[fid], nextPos) = _dec.readBinary(data[offset:])
        elif ftype == 9 or ftype == 10:
            (vtype, vsize, vlen) = _dec.readCollectionBegin(data[offset:])
            offset += vlen
            _data[fid] = []
            _nextPos = 0
            for i in range(vsize):
                if vtype == 12:
                    _aaa, _bbb = self.tryReadThriftContainerStruct(data[offset:], get_data_len=True)
                    _data[fid].append(_aaa)
                    offset += _bbb + 1
        else:
            print(f"[tryReadThriftContainerStruct]不支援Type: {ftype} => ID: {fid}")
        if nextPos > 0:
            data = data[nextPos:]
            c = self.tryReadThriftContainerStruct(data, id=fid, get_data_len=True)
            if c[0] is not None:
                _data.update(c[0])
            nextPos += c[1]
        if get_data_len:
            return [_data, nextPos]
        return _data


# -*- coding: utf-8 -*-
import requests

def loggedIn(func):
    def checkLogin(*args, **kwargs):
        if args[0].can_use_timeline_biz:
            return func(*args, **kwargs)
        else:
            raise Exception("can't use Timeline Biz func")
    return checkLogin

class TimelineBiz():
    can_use_timeline_biz = False

    def __init__(self):
        TIMELINE_BIZ_LIFF_ID = "1654109201-MgN2z4Nd"
        self.can_use_timeline_biz = False
        try:
            self.cmsToken = self.issueLiffView(None, TIMELINE_BIZ_LIFF_ID)[8]
            self.cmsSession = self.getCmsSession()
            self.can_use_timeline_biz = True
        except Exception as e:
            self.log(f"can't use Timeline Biz: {e}")

    def getCmsSession(self):
        url = 'https://timeline.line.biz/api/auth/getSessionByIdToken?idToken=' + self.cmsToken
        req = requests.session()
        res = req.get(url)
        return req

    @loggedIn
    def getLiffCert(self):
        url = 'https://api.line.me/oauth2/v2.1/certs'
        res = self.cmsSession.get(url, headers={"authorization": f"Bearer {self.cmsToken}"})
        return res.json()

    @loggedIn
    def getCmsUser(self):
        url = 'https://timeline.line.biz/api/cmsUser'
        res = self.cmsSession.get(url)
        return res.json()

    @loggedIn
    def getOAList(self):
        url = 'https://timeline.line.biz/api/timeline/v2/bot/list'
        res = self.cmsSession.get(url)
        return res.json()

    @loggedIn
    def createOA(self, displayName, count=100):
        url = 'https://timeline.line.biz/api/timeline/v2/bot/create'
        data = {
            "displayName": displayName,
            "category": 189, 
            "statusMessage":"台灣台灣台灣灣鴻妳好帥台灣台灣台灣台灣"
        }
        headers = {
            'referer': 'https://timeline.line.biz/liff/liff-account/information-terms',
            'origin': 'https://timeline.line.biz',
            'user-agent': 'Mozilla/5.0 (Linux; Android 9.0.0; SOV38 Build/NMF26X; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/74.0.3729.136 Mobile Safari/537.36 Line/12.1.0-dev LIFF',
            'x-botcms-lastclickedelement': "",
            'x-botcms-scriptrevision': '1.0.0-rc',
            'x-requested-with': 'jp.naver.line.android',
            "x-xsrf-token": self.cmsSession.cookies['XSRF-TOKEN'],
            'content-type': 'application/json;charset=UTF-8'
        }
        bots = []
        for i in range(count):
            data['displayName'] = f"{displayName} - {i + 1}"
            res = self.cmsSession.post(url, json=data, headers=headers)
            bots.append(res.json())
        return bots
        # https://timeline.line.biz/api/auth/getSessionByIdToken?idToken= + liff token
        # https://timeline.line.biz/api/common/init
        # https://timeline.line.biz/api/cmsUser
        # https://api.line.me/oauth2/v2.1/verify?access_token=
        # https://timeline.line.biz/api/bots/%40491jpvse/timeline/v2/bot/bridge
        # url = f'https://timeline.line.biz/api/bots/{currentSearchId}/timeline/v2/comment'
        # url2 = f'https://timeline.line.biz/api/bots/{currentSearchId}/timeline/v2/like'
        # url3 = f'https://timeline.line.biz/api/bots/{currentSearchId}/issueInvitationUrl'


# -*- coding: utf-8 -*-

class SecondaryPwlessLoginService(object):
    
    def __init__(self):
        pass

    def createPwlessSession(self, phone, region='TW'):
        params = [
            [12, 1, [
                [11, 1, phone],
                [11, 2, region]
            ]]
        ]
        sqrd = self.generateDummyProtocol('createSession', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 4)

    def verifyLoginCertificate(self, session, cert=None):
        params = [
            [12, 1, [
                [11, 1, session],
                [11, 2, cert]
            ]]
        ]
        sqrd = self.generateDummyProtocol('verifyLoginCertificate', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 4)

    def requestPinCodeVerif(self, session):
        params = [
            [12, 1, [
                [11, 1, session]
            ]]
        ]
        sqrd = self.generateDummyProtocol('requestPinCodeVerif', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 4)

    def putExchangeKey(self, session, temporalPublicKey, e2eeVersion=1):
        params = [
            [12, 1, [
                [11, 1, session],
                [13, 2, [11, 11, {
                    'e2eeVersion': str(e2eeVersion),
                    'temporalPublicKey': temporalPublicKey
                }]]
            ]]
        ]
        sqrd = self.generateDummyProtocol('putExchangeKey', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 4)

    def requestPaakAuth(self, session):
        params = [
            [12, 1, [
                [11, 1, session]
            ]]
        ]
        sqrd = self.generateDummyProtocol('requestPaakAuth', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 4)

    def getE2eeKey(self, session):
        params = [
            [12, 1, [
                [11, 1, session]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getE2eeKey', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 3)

    def pwlessLogin(self, session):
        params = [
            [12, 1, [
                [11, 1, session],
                [11, 2, "DeachSword-CHRLINE"],
                [11, 3, "CHANNELGW"]
            ]]
        ]
        sqrd = self.generateDummyProtocol('login', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 3)

    def pwlessLoginV2(self, session):
        params = [
            [12, 1, [
                [11, 1, session],
                [2, 2, True],
                [11, 3, "DeachSword-CHRLINE"],
                [11, 4, "CHANNELGW"]
            ]]
        ]
        sqrd = self.generateDummyProtocol('loginV2', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_ENDPOINT ,sqrd, 3)


# -*- coding: utf-8 -*-
import time
import json

class LiffService(object):

    def __init__(self):
        pass
        
    def issueLiffView(self, chatMid, liffId="1562242036-RW04okm", lang='zh_TW', deviceSetting=None):
        b = self.TCompactProtocol()
        sqrd = [130, 33, 00] + self.getStringBytes('issueLiffView', isCompact=True)
        sqrd += b.getFieldHeader(12, 1)
        c = self.TCompactProtocol()
        sqrd += c.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(liffId, isCompact=True)
        sqrd += c.getFieldHeader(12, 2)
        d = self.TCompactProtocol()
        if chatMid is None:
            sqrd += d.getFieldHeader(12, 1)
        else:
            if chatMid[0] not in ['u', 'c', 'r']:
                sqrd += d.getFieldHeader(12, 3)
            else:
                sqrd += d.getFieldHeader(12, 2)
            e = self.TCompactProtocol()
            sqrd += e.getFieldHeader(8, 1)
            sqrd += self.getStringBytes(chatMid, isCompact=True)
        sqrd += [0, 0, 0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_LIFF_ENDPOINT ,sqrd, ttype=4)
        
    def getLiffViewWithoutUserContext(self, liffId="1562242036-RW04okm"):
        b = self.TCompactProtocol()
        sqrd = [130, 33, 00] + self.getStringBytes('getLiffViewWithoutUserContext', isCompact=True)
        sqrd += b.getFieldHeader(12, 1)
        c = self.TCompactProtocol()
        sqrd += c.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(liffId, isCompact=True)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_LIFF_ENDPOINT ,sqrd, ttype=4)
        
    def issueSubLiffView(self, chatMid, msit, liffId="1562242036-RW04okm", lang='zh_TW', deviceSetting=None):
        b = self.TCompactProtocol()
        sqrd = [130, 33, 00] + self.getStringBytes('issueSubLiffView', isCompact=True)
        sqrd += b.getFieldHeader(12, 1)
        c = self.TCompactProtocol()
        sqrd += c.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(liffId, isCompact=True)
        sqrd += c.getFieldHeader(8, 3) + self.getStringBytes(lang, isCompact=True)
        sqrd += c.getFieldHeader(8, 5) + self.getStringBytes(msit, isCompact=True)
        sqrd += c.getFieldHeader(12, 2)
        d = self.TCompactProtocol()
        if chatMid[0] not in ['u', 'c', 'r']:
            sqrd += d.getFieldHeader(12, 3)
        else:
            sqrd += d.getFieldHeader(12, 2)
        e = self.TCompactProtocol()
        sqrd += e.getFieldHeader(8, 1)
        sqrd += self.getStringBytes(chatMid, isCompact=True)
        sqrd += [0, 0, 0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_LIFF_ENDPOINT ,sqrd, ttype=4)


# -*- coding: utf-8 -*-
import time
import json

class ShopService(object):

    def __init__(self):
        pass
        
    def getProduct(self, shopId, productId, language="zh-TW", country="TW"):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 10, 103, 101, 116, 80, 114, 111, 100, 117, 99, 116, 0, 0, 0, 0]
        sqrd += [11, 0, 2, 0, 0, 0, len(shopId)] # e.g. stickershop
        for value in shopId:
            sqrd.append(ord(value))
        sqrd += [11, 0, 3, 0, 0, 0, len(productId)]
        for value in productId:
            sqrd.append(ord(value))
        sqrd += [12, 0, 4]
        sqrd += [11, 0, 1, 0, 0, 0, len(language)]
        for value in language:
            sqrd.append(ord(value))
        sqrd += [11, 0, 2, 0, 0, 0, len(country)]
        for value in country:
            sqrd.append(ord(value))
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_UNIFIED_SHOP_ENDPOINT ,sqrd)

    def getProductsByAuthor(self, authorId, productType=1):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getProductsByAuthor') + [0, 0, 0, 0]
        sqrd += [12, 0, 2]
        sqrd += [8, 0, 1] + self.getIntBytes(productType)
        sqrd += [11, 0, 2] + self.getStringBytes(authorId)
        sqrd += [8, 0, 3] + self.getIntBytes(0)
        sqrd += [8, 0, 4] + self.getIntBytes(100)
        sqrd += [2, 0, 6, int(True)]
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_UNIFIED_SHOP_ENDPOINT ,sqrd)
    
    def getStudentInformation(self):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getStudentInformation') + [0, 0, 0, 0]
        sqrd += [12, 0, 2]
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_UNIFIED_SHOP_ENDPOINT ,sqrd)
    
    def canReceivePresent(self, shopId, productId, recipientMid):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('canReceivePresent') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(shopId)
        sqrd += [11, 0, 3] + self.getStringBytes(productId)
        sqrd += [12, 0, 4]
        sqrd += [11, 0, 1] + self.getStringBytes('zh_TW') #language
        sqrd += [11, 0, 2] + self.getStringBytes('TW') #country
        sqrd += [0]
        sqrd += [11, 0, 5] + self.getStringBytes(recipientMid)
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_UNIFIED_SHOP_ENDPOINT ,sqrd)
    
    def getOwnedProductSummaries(self, shopId, offset=0, limit=200, language='zh_TW', country='TW'):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getOwnedProductSummaries') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(shopId)
        sqrd += [8, 0, 3] + self.getIntBytes(offset)
        sqrd += [8, 0, 4] + self.getIntBytes(limit)
        sqrd += [12, 0, 5]
        sqrd += [11, 0, 1] + self.getStringBytes(language)
        sqrd += [11, 0, 2] + self.getStringBytes(country)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_UNIFIED_SHOP_ENDPOINT ,sqrd)


# -*- coding: utf-8 -*-

class ChannelService(object):

    def __init__(self):
        pass
        
    def issueChannelToken(self, channelId="1341209950"):
        #sqrd = self.DummyProtocol("issueChannelToken", 3, {
        #    1: (11, channelId)
        #}).read()
        sqrd = [128, 1, 0, 1] + self.getStringBytes('issueChannelToken') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(channelId)
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_CHANNEL_ENDPOINT ,sqrd)
        
    def approveChannelAndIssueChannelToken(self, channelId="1341209950"):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('approveChannelAndIssueChannelToken') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(channelId)
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_CHANNEL_ENDPOINT ,sqrd)
        
    def getChannelInfo(self, channelId):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 14, 103, 101, 116, 67, 104, 97, 110, 110, 101, 108, 73, 110, 102, 111, 0, 0, 0, 0, 11, 0, 2, 0, 0, 0, len(channelId)]
        for value in str(channelId):
            sqrd.append(ord(value))
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_CHANNEL_ENDPOINT ,sqrd)
        
    def getCommonDomains(self, lastSynced=0):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 16, 103, 101, 116, 67, 111, 109, 109, 111, 110, 68, 111, 109, 97, 105, 110, 115, 0, 0, 0, 0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_CHANNEL_ENDPOINT ,sqrd)
        
    def issueRequestTokenWithAuthScheme(self, channelId, otpId, authScheme, returnUrl):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 31, 105, 115, 115, 117, 101, 82, 101, 113, 117, 101, 115, 116, 84, 111, 107, 101, 110, 87, 105, 116, 104, 65, 117, 116, 104, 83, 99, 104, 101, 109, 101, 0, 0, 0, 0]
        sqrd += [11, 0, 1, 0, 0, 0, len(channelId)]
        for value in channelId:
            sqrd.append(ord(value))
        sqrd += [11, 0, 2, 0, 0, 0, len(otpId)]
        for value in otpId:
            sqrd.append(ord(value))
        sqrd += [15, 0, 3, 11, 0, 0, 0, len(authScheme)]
        for mid in authScheme:
            sqrd += [0, 0, 0, len(mid)]
            for value in mid:
                sqrd.append(ord(value))
        sqrd += [11, 0, 4, 0, 0, 0, len(returnUrl)]
        for value in returnUrl:
            sqrd.append(ord(value))
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_CHANNEL_ENDPOINT ,sqrd)
    
    def getReturnUrlWithRequestTokenForAutoLogin(self, url):
        params = [
            [12, 2, [
                [11, 1, url]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getReturnUrlWithRequestTokenForAutoLogin', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CHANNEL_ENDPOINT_V4 ,sqrd, 4)


# -*- coding: utf-8 -*-

class SquareBotService(object):

    def __init__(self):
        pass
        
    def getSquareBot(self, botMid):
        params = [
            [12, 1, [
                [11, 1, botMid]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getSquareBot', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_BOT_ENDPOINT ,sqrd)


# -*- coding: utf-8 -*-

class CallService(object):

    def __init__(self):
        pass
        
    def acquireCallRoute(self, to, callType=1, fromEnvInfo={}):
        params = [
            [11, 2, to],
            [8, 3, callType],
            [13, 4, [11, 11, fromEnvInfo]]
        ]
        sqrd = self.generateDummyProtocol('acquireCallRoute', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def acquireOACallRoute(self, searchId, fromEnvInfo, otp):
        params = [
            [12, 2, [
                [11, 1, searchId],
                [13, 2, [11, 11, fromEnvInfo]],
                [11, 3, otp]
            ]]
        ]
        sqrd = self.generateDummyProtocol('acquireOACallRoute', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def searchPaidCallUserRate(self, countryCode):
        params = [
            [11, 2, countryCode],
        ]
        sqrd = self.generateDummyProtocol('searchPaidCallUserRate', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def acquirePaidCallCurrencyExchangeRate(self, countryCode):
        params = [
            [11, 2, countryCode],
        ]
        sqrd = self.generateDummyProtocol('acquirePaidCallCurrencyExchangeRate', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def lookupPaidCall(self, dialedNumber, language, referer):
        params = [
            [11, 2, dialedNumber],
            [11, 3, language],
            [11, 4, referer],
        ]
        sqrd = self.generateDummyProtocol('lookupPaidCall', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def acquirePaidCallRoute(self, paidCallType, dialedNumber, language, networkCode, disableCallerId, referer, adSessionId):
        params = [
            [8, 2, paidCallType],
            [11, 3, dialedNumber],
            [11, 4, language],
            [11, 5, networkCode],
            [2, 6, disableCallerId],
            [11, 7, referer],
            [11, 8, adSessionId],
        ]
        sqrd = self.generateDummyProtocol('acquirePaidCallRoute', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getPaidCallBalanceList(self, language):
        params = [
            [11, 2, language],
        ]
        sqrd = self.generateDummyProtocol('getPaidCallBalanceList', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getPaidCallHistory(self, start, size):
        params = [
            [10, 2, start],
            [8, 3, size],
        ]
        sqrd = self.generateDummyProtocol('getPaidCallHistory', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getCallCreditProducts(self, appStoreCode, pgCode, country):
        params = [
            [8, 2, appStoreCode],
            [8, 3, pgCode],
            [11, 4, country],
        ]
        sqrd = self.generateDummyProtocol('getCallCreditProducts', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def reserveCallCreditPurchase(self, productId, country, currency, price, appStoreCode, language, pgCode, redirectUrl):
        params = [
            [12, 2, [
                [11, 1, productId],
                [11, 2, country],
                [11, 3, currency],
                [11, 4, price],
                [8, 5, appStoreCode],
                [11, 6, language],
                [8, 7, pgCode],
                [11, 8, redirectUrl],
            ]]
        ]
        sqrd = self.generateDummyProtocol('reserveCallCreditPurchase', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getCallCreditPurchaseHistory(self, start, size, language, eddt, appStoreCode):
        params = [
            [12, 2, [
                [10, 1, start],
                [8, 2, size],
                [11, 3, language],
                [11, 4, eddt],
                [8, 5, appStoreCode],
            ]]
        ]
        sqrd = self.generateDummyProtocol('getCallCreditPurchaseHistory', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def redeemPaidCallVoucher(self, serial):
        params = [
            [11, 2, serial]
        ]
        sqrd = self.generateDummyProtocol('redeemPaidCallVoucher', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getPaidCallMetadata(self, language):
        params = [
            [11, 2, language],
        ]
        sqrd = self.generateDummyProtocol('getPaidCallMetadata', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def acquireGroupCallRoute(self, chatMid, mediaType=1, isInitialHost=True, capabilities=[]):
        params = [
            [11, 2, chatMid],
            [8, 3, mediaType],
            [2, 4, isInitialHost],
            [15, 5, [11, capabilities]],
        ]
        sqrd = self.generateDummyProtocol('acquireGroupCallRoute', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getGroupCall(self, chatMid):
        params = [
            [11, 2, chatMid],
        ]
        sqrd = self.generateDummyProtocol('getGroupCall', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def inviteIntoGroupCall(self, chatMid, memberMids, mediaType=1):
        params = [
            [11, 2, chatMid],
            [15, 3, [11, memberMids]],
            [8, 4, mediaType],
        ]
        sqrd = self.generateDummyProtocol('inviteIntoGroupCall', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def markPaidCallAd(self, dialedNumber, language, disableCallerId):
        params = [
            [11, 2, dialedNumber],
            [11, 3, language],
            [2, 4, disableCallerId],
        ]
        sqrd = self.generateDummyProtocol('markPaidCallAd', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getPaidCallAdStatus(self, dialedNumber, language, disableCallerId):
        params = []
        sqrd = self.generateDummyProtocol('getPaidCallAdStatus', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def acquireTestCallRoute(self, dialedNumber, language, disableCallerId):
        params = []
        sqrd = self.generateDummyProtocol('acquireTestCallRoute', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getGroupCallUrls(self):
        params = [
            [12, 2, []]
        ]
        sqrd = self.generateDummyProtocol('getGroupCallUrls', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def createGroupCallUrl(self, title):
        params = [
            [12, 2, [
                [11, 1, title]
            ]]
        ]
        sqrd = self.generateDummyProtocol('createGroupCallUrl', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def deleteGroupCallUrl(self, urlId):
        params = [
            [12, 2, [
                [11, 1, urlId]
            ]]
        ]
        sqrd = self.generateDummyProtocol('deleteGroupCallUrl', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def updateGroupCallUrl(self, urlId, title):
        params = [
            [12, 2, [
                [11, 1, urlId],
                [12, 2, [
                    [11, 1, title]
                ]]
            ]]
        ]
        sqrd = self.generateDummyProtocol('updateGroupCallUrl', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def getGroupCallUrlInfo(self, urlId):
        params = [
            [12, 2, [
                [11, 1, urlId]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getGroupCallUrlInfo', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)

    def joinChatByCallUrl(self, urlId):
        params = [
            [12, 2, [
                [11, 1, urlId],
                [8, 2, self.getCurrReqId()]
            ]]
        ]
        sqrd = self.generateDummyProtocol('joinChatByCallUrl', params, 3)
        return self.postPackDataAndGetUnpackRespData(self.LINE_CALL_ENDPOINT ,sqrd, 3)


# -*- coding: utf-8 -*-

class SettingsService(object):
    
    def __init__(self):
        pass
    
    def getSetting(self):
        params = []
        sqrd = self.generateDummyProtocol('getSetting', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def contextAgnosticGetSetting(self):
        params = []
        sqrd = self.generateDummyProtocol('contextAgnosticGetSetting', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def setSetting(self):
        params = []
        sqrd = self.generateDummyProtocol('setSetting', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def setSettingWithScope(self):
        params = []
        sqrd = self.generateDummyProtocol('setSettingWithScope', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def resetSetting(self):
        params = []
        sqrd = self.generateDummyProtocol('resetSetting', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def resetSettingWithScope(self):
        params = []
        sqrd = self.generateDummyProtocol('resetSettingWithScope', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def searchSettings(self):
        params = []
        sqrd = self.generateDummyProtocol('searchSettings', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)

    def contextAgnosticSearchSettings(self):
        params = []
        sqrd = self.generateDummyProtocol('contextAgnosticSearchSettings', params, 4)
        return self.postPackDataAndGetUnpackRespData("/US4" ,sqrd, 4)


# -*- coding: utf-8 -*-
import time
import json
import uuid

class PrimaryAccountInitService(object):

    def __init__(self):
        self.register_headers  = {
            "x-line-application": "ANDROID\t11.14.0\tAndroid OS\t10.0.1",
            "x-le": self.le,
            "x-lcs": self._encryptKey,
            "User-Agent": "Line/11.14.0",
            "content-type": "application/x-thrift; protocol=TBINARY",
            "x-lpv": "1",
        }
        self.uuid = uuid.uuid4().hex
    
    def setPrimaryUuid(self, uuid):
        self.uuid = uuid
        
    def openPrimarySession(self):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('openSession') + [0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [13, 0, 1, 11, 11] + self.getIntBytes(0)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def getCountryInfo(self, authSessionId, simCard=None):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getCountryInfo') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def getPhoneVerifMethod(self, authSessionId, phoneNumber, countryCode, deviceModel="SM-N950F"):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getPhoneVerifMethod') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 2]
        sqrd += [11, 0, 1] + self.getStringBytes(self.uuid)
        sqrd += [11, 0, 2] + self.getStringBytes(deviceModel)
        sqrd += [0]
        sqrd += [12, 0, 3]
        sqrd += [11, 0, 1] + self.getStringBytes(phoneNumber)
        sqrd += [11, 0, 2] + self.getStringBytes(countryCode)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def sendPinCodeForPhone(self, authSessionId, phoneNumber, countryCode, deviceModel="SM-N950F"):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('sendPinCodeForPhone') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 2]
        sqrd += [11, 0, 1] + self.getStringBytes(self.uuid)
        print(f"UUID: {self.uuid}")
        sqrd += [11, 0, 2] + self.getStringBytes(deviceModel)
        sqrd += [0]
        sqrd += [12, 0, 3]
        sqrd += [11, 0, 1] + self.getStringBytes(phoneNumber)
        sqrd += [11, 0, 2] + self.getStringBytes(countryCode)
        sqrd += [0]
        sqrd += [8, 0, 4] + self.getIntBytes(2)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def verifyPhone(self, authSessionId, phoneNumber, countryCode, pinCode, deviceModel="SM-N950F"):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('verifyPhone') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 2]
        sqrd += [11, 0, 1] + self.getStringBytes(self.uuid)
        sqrd += [11, 0, 2] + self.getStringBytes(deviceModel)
        sqrd += [0]
        sqrd += [12, 0, 3]
        sqrd += [11, 0, 1] + self.getStringBytes(phoneNumber)
        sqrd += [11, 0, 2] + self.getStringBytes(countryCode)
        sqrd += [0]
        sqrd += [11, 0, 4] + self.getStringBytes(pinCode)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def validateProfile(self, authSessionId, displayName):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('validateProfile') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [11, 0, 2] + self.getStringBytes(displayName)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def exchangeEncryptionKey(self, authSessionId, publicKey, nonce, authKeyVersion=1):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('exchangeEncryptionKey') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 2]
        sqrd += [8, 0, 1] + self.getIntBytes(authKeyVersion)
        sqrd += [11, 0, 2] + self.getStringBytes(publicKey)
        sqrd += [11, 0, 3] + self.getStringBytes(nonce)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def setPassword(self, authSessionId, cipherText, encryptionKeyVersion=1):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('setPassword') + [0, 0, 0, 0]
        sqrd += [11, 0, 1] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 2]
        sqrd += [8, 0, 1] + self.getIntBytes(encryptionKeyVersion)
        sqrd += [11, 0, 2] + self.getStringBytes(cipherText)
        sqrd += [0, 0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def registerPrimaryUsingPhone(self, authSessionId):
        _headers = {
            'x-lpqs': "/acct/pais/v1"
        }
        a = self.encHeaders(_headers)
        sqrd = [128, 1, 0, 1] + self.getStringBytes('registerPrimaryUsingPhone') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [0]
        sqr_rd = a + sqrd
        _data = bytes(sqr_rd)
        data = self.encData(_data)
        res = self.server.postContent('https://ga2.line.naver.jp', data=data, headers=self.register_headers)
        data = self.decData(res.content)
        return self.tryReadData(data)
        
    def getPhoneVerifMethodV2(self, authSessionId, phoneNumber, countryCode, deviceModel="SM-N950F"):
        params = [
            [12, 1, [
                [11, 1, authSessionId],
                [12, 2, [
                    [11, 1, self.uuid],
                    [11, 2, deviceModel]
                ]],
                [12, 3, [
                    [11, 1, phoneNumber],
                    [11, 2, countryCode]
                ]],
            ]]
        ]
        sqrd = self.generateDummyProtocol('getPhoneVerifMethodV2', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/pais/v1" ,sqrd, 3, headers=self.register_headers)
        
    def requestToSendPhonePinCode(self, authSessionId, phoneNumber, countryCode, verifMethod=1):
        params = [
            [12, 1, [
                [11, 1, authSessionId],
                [12, 2, [
                    [11, 1, phoneNumber],
                    [11, 2, countryCode]
                ]],
                [8, 3, verifMethod],
            ]]
        ]
        sqrd = self.generateDummyProtocol('requestToSendPhonePinCode', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/pais/v1" ,sqrd, 3, headers=self.register_headers)
        
    def verifyPhonePinCode(self, authSessionId, phoneNumber, countryCode, pinCode):
        params = [
            [12, 1, [
                [11, 1, authSessionId],
                [12, 2, [
                    [11, 1, phoneNumber],
                    [11, 2, countryCode]
                ]],
                [11, 3, pinCode],
            ]]
        ]
        sqrd = self.generateDummyProtocol('verifyPhonePinCode', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/pais/v1" ,sqrd, 3, headers=self.register_headers)
        
    def verifyAccountUsingPwd(self, authSessionId, identifier, countryCode, cipherText):
        params = [
            [12, 1, [
                [11, 1, authSessionId],
                [12, 2, [
                    [8, 1, 1], # type
                    [11, 2, identifier],
                    [11, 3, countryCode]
                ]],
                [12, 3, [
                    [8, 1, 1], # encryptionKeyVersion
                    [11, 2, cipherText]
                ]],
            ]]
        ]
        sqrd = self.generateDummyProtocol('verifyAccountUsingPwd', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/pais/v1" ,sqrd, 3, headers=self.register_headers)
        
    def registerPrimaryUsingPhoneWithTokenV3(self, authSessionId):
        params = [
            [11, 2, authSessionId]
        ]
        sqrd = self.generateDummyProtocol('registerPrimaryUsingPhoneWithTokenV3', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/pais/v1" ,sqrd, 3, headers=self.register_headers)
        
    def registerPrimaryWithTokenV3(self, authSessionId):
        params = [
            [11, 2, authSessionId]
        ]
        sqrd = self.generateDummyProtocol('registerPrimaryUsingPhoneWithTokenV3', params, 3)
        return self.postPackDataAndGetUnpackRespData("/acct/pais/v1" ,sqrd, 3, headers=self.register_headers)


# -*- coding: utf-8 -*-
import time
import json
import requests

class AuthService(object):
    
    def __init__(self):
        pass
        
    def openAuthSession(self):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('openAuthSession') + [0, 0, 0, 0]
        sqrd += [12, 0, 2]
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)
        
    def getAuthRSAKey(self, authSessionId: str):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('getAuthRSAKey') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [8, 0, 3] + self.getIntBytes(1) #identityProvider
        sqrd += [0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)

    def setIdentifier(self, authSessionId: str, cipherKeyId: str, cipherText: str):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('setIdentifier') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 3]
        sqrd += [8, 0, 2] + self.getIntBytes(1) #identityProvider
        sqrd += [11, 0, 3] + self.getStringBytes(cipherKeyId) #cipherKeyId, eg.10031
        sqrd += [11, 0, 4] + self.getStringBytes(cipherText) #cipherText, rsa enc maybe
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)

    def updateIdentifier(self, authSessionId: str, cipherKeyId: str, cipherText: str):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('updateIdentifier') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 3]
        sqrd += [8, 0, 2] + self.getIntBytes(1) #identityProvider
        sqrd += [11, 0, 3] + self.getStringBytes(cipherKeyId) #cipherKeyId, eg.10031
        sqrd += [11, 0, 4] + self.getStringBytes(cipherText) #cipherText, rsa enc maybe
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)
        
    def resendIdentifierConfirmation(self, authSessionId: str):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('resendIdentifierConfirmation') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 3]
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)
        
    def confirmIdentifier(self, authSessionId: str, verificationCode: str):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('confirmIdentifier') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 3]
        sqrd += [12, 0, 5]
        # sqrd += [2, 0, 2] + self.getIntBytes(1) #forceRegistration : 1?
        sqrd += [11, 0, 3] + self.getStringBytes(verificationCode) #verificationCode
        sqrd += [0, 0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)
        
    def removeIdentifier(self, authSessionId: str, cipherKeyId: str, cipherText: str):
        sqrd = [128, 1, 0, 1] + self.getStringBytes('removeIdentifier') + [0, 0, 0, 0]
        sqrd += [11, 0, 2] + self.getStringBytes(authSessionId)
        sqrd += [12, 0, 3]
        sqrd += [8, 0, 2] + self.getIntBytes(1) #identityProvider
        sqrd += [11, 0, 3] + self.getStringBytes(cipherKeyId) #cipherKeyId
        sqrd += [11, 0, 4] + self.getStringBytes(cipherText) #cipherText
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT ,sqrd)
        
    def getClovaAppToken(self, authSessionId, cipherText, metaData={}):
        params = [
            [11, 2, authSessionId],
            [12, 3, [
                [8, 1, 2],
                [13, 2, [11, 11, metaData]],
                [11, 3, cipherText]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getClovaAppToken', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT_V4 ,sqrd, 4)
        
    def loginFromClova(self, authSessionId, cipherText, metaData={}):
        params = [
            [11, 2, authSessionId],
            [12, 3, [
                [8, 1, 2],
                [13, 2, [11, 11, metaData]],
                [11, 3, cipherText]
            ]]
        ]
        sqrd = self.generateDummyProtocol('loginFromClova', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT_V4 ,sqrd, 4)
        
    def validateClovaRequest(self, authSessionId, cipherText, metaData={}):
        params = [
            [11, 2, authSessionId],
            [12, 3, [
                [8, 1, 2],
                [13, 2, [11, 11, metaData]],
                [11, 3, cipherText]
            ]]
        ]
        sqrd = self.generateDummyProtocol('validateClovaRequest', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT_V4 ,sqrd, 4)
        
    def setClovaCredential(self, authSessionId, cipherText, metaData={}):
        params = [
            [11, 2, authSessionId],
            [12, 3, [
                [8, 1, 2], #authLoginVersion
                [13, 2, [11, 11, metaData]],
                [11, 3, cipherText]
            ]]
        ]
        sqrd = self.generateDummyProtocol('setClovaCredential', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT_V4 ,sqrd, 4)
        
    def validateClovaAppToken(self, authSessionId, cipherText, metaData={}):
        params = [
            [11, 2, authSessionId],
            [12, 3, [
                [8, 1, 2], #authLoginVersion
                [13, 2, [11, 11, metaData]],
                [11, 3, cipherText]
            ]]
        ]
        sqrd = self.generateDummyProtocol('validateClovaAppToken', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_AUTH_ENDPOINT_V4 ,sqrd, 4)
        
    def verifyQrcodeWithE2EE(self, verifier, keyId, keyData, createdTime, encryptedKeyChain, hashKeyChain, pinCode=""):
        params = [
            [11, 2, verifier],
            [11, 3, pinCode],
            [8, 4, 95], # errorCode
            [12, 5, [
                [8, 1, 1], # version
                [8, 2, keyId],
                [11, 4, keyData],
                [10, 5, createdTime]
            ]],
            [11, 6, encryptedKeyChain],
            [11, 7, hashKeyChain]
        ]
        sqrd = self.generateDummyProtocol('verifyQrcodeWithE2EE', params, 4)
        return self.postPackDataAndGetUnpackRespData("/S4" ,sqrd, 4)
        
    def confirmE2EELogin(self, verifier, deviceSecret):
        params = [
            [11, 1, verifier],
            [11, 2, deviceSecret],
        ]
        sqrd = self.generateDummyProtocol('confirmE2EELogin', params, 3)
        return self.postPackDataAndGetUnpackRespData("/api/v3p/rs" ,sqrd, 3)


"""
Author: YinMo
Version: 0.0.1-beta
Description: died
"""
import hashlib, json, os
import axolotl_curve25519 as Curve25519
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from Crypto.Cipher import AES

class E2EE():

    def generateSharedSecret(self, private_key, public_key):
        return Curve25519.calculateAgreement(bytes(private_key), bytes(public_key))
    
    def _xor(self, buf):
        buf_length = int(len(buf) / 2)
        buf2 = bytearray(buf_length)
        for i in range(buf_length):
            buf2[i] = buf[i] ^ buf[buf_length + i]
        return bytes(buf2)
    
    def getSHA256Sum(self, *args):
        instance = hashlib.sha256()
        for arg in args:
            if isinstance(arg, str):
                arg = arg.encode()
            instance.update(arg)
        return instance.digest()
    
    def _encryptAESECB(self, aes_key, plain_data):
        aes = AES.new(aes_key, AES.MODE_ECB)
        return aes.encrypt(plain_data)
    
    def decryptKeyChain(self, publicKey, privateKey, encryptedKeyChain):
        shared_secret = self.generateSharedSecret(privateKey, publicKey)
        aes_key = self.getSHA256Sum(shared_secret, 'Key')
        aes_iv = self._xor(self.getSHA256Sum(shared_secret, 'IV'))
        aes = AES.new(aes_key, AES.MODE_CBC, aes_iv)
        keychain_data = aes.decrypt(encryptedKeyChain)
        key = keychain_data.hex()
        key = bin2bytes(key)
        key = self.tryReadThriftContainerStruct(key)
        public_key = bytes(key[1][0][4])
        private_key = bytes(key[1][0][5])
        return [private_key, public_key]
    
    def encryptDeviceSecret (self, publicKey, privateKey, encryptedKeyChain):
        shared_secret = self.generateSharedSecret(privateKey, publicKey)
        aes_key = self.getSHA256Sum(shared_secret, 'Key')
        encryptedKeyChain = self._xor(self.getSHA256Sum(encryptedKeyChain))
        keychain_data = self._encryptAESECB(aes_key, encryptedKeyChain)
        return keychain_data
    
    def generateAAD(self, a, b, c, d, e=2, f=0):
        aad = b''
        aad += a.encode()
        aad += b.encode()
        aad += bytes(self.getIntBytes(c))
        aad += bytes(self.getIntBytes(d))
        aad += bytes(self.getIntBytes(e))
        aad += bytes(self.getIntBytes(f))
        return aad
    
    def encryptE2EETextMessage(self, senderKeyId, receiverKeyId, keyData, specVersion, text, to ,_from):
        #selfKey = self.getE2EEKeys(self.mid)
        salt = os.urandom(16)
        gcmKey = self.getSHA256Sum(keyData, salt, b'Key')
        gcmIV = self.getSHA256Sum(keyData, salt, b'IV')
        aad = self.generateAAD(to, _from, senderKeyId, receiverKeyId, specVersion, 0)
        sign = os.urandom(16)
        data = json.dumps({
            'text': text
        }).encode()
        encData = self.encryptE2EEMessageV2(data, gcmKey, sign, aad)
        self.log(f'senderKeyId: {senderKeyId} ({self.getIntBytes(senderKeyId)})', True)
        self.log(f'receiverKeyId: {receiverKeyId} ({self.getIntBytes(receiverKeyId)})', True)
        return [salt, encData, sign, bytes(self.getIntBytes(senderKeyId)), bytes(self.getIntBytes(receiverKeyId))]
    
    def decryptE2EETextMessage(self, messageObj):
        chunks = messageObj[20]
        salt = chunks[0]
        message = chunks[1]
        sign = chunks[2]
        senderKeyId = __byte2int(chunks[3])
        receiverKeyId = __byte2int(chunks[4])
        
        _key = self.negotiateE2EEPublicKey(messageObj[0]) # todo: to or _from
        
        aesKey = self.generateSharedSecret(self.getPrivateKey(self.mid), _key[2][4])
        gcmKey = self.getSHA256Sum(aesKey, salt, b'Key')
        s = hashlib.sha256()
        s.update(aesKey)
        s.update(salt)
        s.update(b'IV')
        iv = s.digest()
        aad = self.generateAAD(message[0], message[1], senderKeyId, receiverKeyId)
        
        aesgcm = AESGCM(gcmKey)
        decrypted = aesgcm.decrypt(sign, message, aad)
        self.log(f'decrypted: {decrypted}', True)
        return json.loads(decrypted)['text'] # todo: contentType
    
    def decryptE2EETextMessageV2(self, to , _from, chunks, privK, pubK):
        for i in range(len(chunks)):
            if isinstance(chunks[i], str):
                chunks[i] = chunks[i].encode()
        salt = chunks[0]
        message = chunks[1]
        sign = chunks[2]
        senderKeyId = byte2int(chunks[3])
        receiverKeyId = byte2int(chunks[4])
        self.log(f'senderKeyId: {senderKeyId}', True)
        self.log(f'receiverKeyId: {receiverKeyId}', True)
        
        aesKey = self.generateSharedSecret(privK, pubK)
        gcmKey = self.getSHA256Sum(aesKey, salt, b'Key')
        iv = self.getSHA256Sum(aesKey, salt, b'IV')
        aad = self.generateAAD(to, _from, senderKeyId, receiverKeyId)
        
        aesgcm = AESGCM(gcmKey)
        decrypted = aesgcm.decrypt(sign, message, aad)
        self.log(f'decrypted: {decrypted}', True)
        return json.loads(decrypted)['text']
    
    def encryptE2EEMessageV2(self, data, gcmKey, nonce, aad):
        aesgcm = AESGCM(gcmKey)
        return aesgcm.encrypt(nonce, data, aad)

def byte2int(t):
    e = 0
    i = 0
    s = len(t)
    for i in range(s):
        e = 256 * e + t[i]
    return e

def bin2bytes(k):
    e = []
    for i in range(int(len(k) / 2)):
        _i = int(k[i * 2:i * 2 + 2], 16)
        e.append(_i)
    return bytearray(e)


# -*- coding: utf-8 -*-

class AccessTokenRefreshService(object):
    
    def __init__(self):
        pass

    def refreshAccessToken(self, refreshToken):
        params = [
            [12, 1, [
                [11, 1, refreshToken]
            ]]
        ]
        sqrd = self.generateDummyProtocol('refresh', params, 3)
        return self.postPackDataAndGetUnpackRespData("/EXT/auth/tokenrefresh/v1" ,sqrd, 3)


# -*- coding: utf-8 -*-
import os, sys, threading, time, traceback

class Poll(object):
    opFuncs = {}

    def __init__(self):
        pass
    def __fetchOps(self, revision, count=100):
        return self.fetchOps(revision, count)
    
    def __execute(self, op, func):
        try:
            func(op, self)
        except Exception as e:
            self.log(traceback.format_exc())

    def addOpFunc(self, opType, func):
        self.opFuncs[opType] = func
    
    def setRevision(self, revision):
        self.revision = max(revision, self.revision)

    def trace(self, func, isThreading=True):
        while self.is_login:
            ops = self.__fetchOps(self.revision)
            if 'error' in ops:
                raise Exception(ops['error'])
            for op in ops:
                if op[3] != 0 and op[3] != -1:
                    self.setRevision(op[1])
                    if isThreading:
                        _td = threading.Thread(target=self.__execute, args=(op, func))
                        _td.daemon = True
                        _td.start()
                    else:
                        self.__execute(op, func)


# -- coding utf-8 --

class SearchService(object)
    
    def __init__(self)
        pass
    
    def searchAll(self)
        params = []
        sqrd = self.generateDummyProtocol('searchAll', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def searchCollection(self)
        params = []
        sqrd = self.generateDummyProtocol('searchCollection', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def searchLineat(self)
        params = []
        sqrd = self.generateDummyProtocol('searchLineat', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def searchByPopularCategory(self)
        params = []
        sqrd = self.generateDummyProtocol('searchByPopularCategory', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def searchByCategory(self)
        params = []
        sqrd = self.generateDummyProtocol('searchByCategory', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def getPopularCategory(self)
        params = []
        sqrd = self.generateDummyProtocol('getPopularCategory', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def getNotice(self)
        params = []
        sqrd = self.generateDummyProtocol('getNotice', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def getSearchSection(self)
        params = []
        sqrd = self.generateDummyProtocol('getSearchSection', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)

    def getAutocomplete(self)
        params = []
        sqrd = self.generateDummyProtocol('getAutocomplete', params, 4)
        return self.postPackDataAndGetUnpackRespData("/search/v3" ,sqrd, 4)


import re, setuptools

with open('CHRLINE/__init__.py') as f:
    version = re.search(r'__version__\s*=\s*\'(.+?)\'', f.read()).group(1)
    
with open("README.md","r") as f:
  long_description = f.read()

setuptools.setup(
  name="CHRLINE",
  version=version,
  author="DeachSword",
  description="LINE API!",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/DeachSword/CHRLINE",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3.6"],
  install_requires=[
    'pycryptodome==3.9.8', # ez:D
    'xxhash',
    'rsa',
    'requests',
    'python-axolotl-curve25519',
    'httpx[http2]',
    'h2>=3.2.0',
    'gevent',
    'cryptography',
])


# -*- coding: utf-8 -*-

class SecondaryPwlessLoginPermitNoticeService(object):
    
    def __init__(self):
        pass

    def checkPwlessPinCodeVerified(self, session):
        params = [
            [12, 1, [
                [11, 1, session],
            ]]
        ]
        sqrd = self.generateDummyProtocol('checkPinCodeVerified', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_PERMIT_ENDPOINT ,sqrd, 4, access_token=session)

    def checkPaakAuthenticated(self, session):
        params = [
            [12, 1, [
                [11, 1, session],
                [11, 2, 'CHANNELGW'],
                [2, 3, True]
            ]]
        ]
        sqrd = self.generateDummyProtocol('checkPaakAuthenticated', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SECONDARY_PWLESS_LOGIN_PERMIT_ENDPOINT ,sqrd, 4, access_token=session)